import datetime
import functools
import typing as t
import uuid
from copy import deepcopy
from unittest.mock import Mock, call, patch

import marshmallow
import pytest
import typesystem

from flama import schemas, types
from flama.injection import Parameter as InjectionParameter
from flama.schemas.data_structures import Field, Parameter, ParameterLocation, Schema
from tests.schemas.test_generator import assert_recursive_contains

Unknown = t.NewType("Unknown", None)


class TestCaseField:
    @pytest.mark.parametrize(
        ["params", "result"],
        (
            pytest.param(
                {"name": "foo", "type": int},
                {
                    "name": "foo",
                    "type": int,
                    "nullable": False,
                    "multiple": False,
                    "required": True,
                    "default": InjectionParameter.empty,
                },
                id="default",
            ),
            pytest.param(
                {"name": "foo", "type": t.Optional[int]},
                {
                    "name": "foo",
                    "type": t.Optional[int],
                    "nullable": True,
                    "multiple": False,
                    "required": True,
                    "default": InjectionParameter.empty,
                },
                id="nullable",
            ),
            pytest.param(
                {"name": "foo", "type": list[int]},
                {
                    "name": "foo",
                    "type": list[int],
                    "nullable": False,
                    "multiple": True,
                    "required": True,
                    "default": InjectionParameter.empty,
                },
                id="multiple",
            ),
        ),
    )
    def test_init(self, params, result):
        field = Field(**params)

        for k, v in result.items():
            assert getattr(field, k) == v

    def test_from_parameter(self):
        assert Field.from_parameter(InjectionParameter(name="foo", annotation=int, default=0)) == Field(
            "foo", int, required=False, default=0
        )

    def test_is_field(self):
        mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            Field.is_field(mock)
            assert schemas_mock.adapter.is_field.call_args_list == [call(mock)]

    @pytest.mark.parametrize(
        ["type_", "result"],
        (
            pytest.param(int, True, id="int"),
            pytest.param(float, True, id="float"),
            pytest.param(str, True, id="str"),
            pytest.param(bool, True, id="bool"),
            pytest.param(uuid.UUID, True, id="uuid"),
            pytest.param(datetime.date, True, id="date"),
            pytest.param(datetime.time, True, id="date"),
            pytest.param(datetime.datetime, True, id="datetime"),
            pytest.param(types.QueryParam, True, id="query_param"),
            pytest.param(types.PathParam, True, id="path_param"),
            pytest.param(t.Optional[int], True, id="nullable"),
            pytest.param(list[int], True, id="list"),
            pytest.param(Mock, False, id="not_valid"),
        ),
    )
    def test_is_http_valid_type(self, type_, result):
        assert Field.is_http_valid_type(type_) == result

    def test_json_schema(self):
        field = Field("foo", int)
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            field.json_schema
            assert schemas_mock.adapter.to_json_schema.call_args_list == [call(field.field)]


class TestCaseSchema:
    @pytest.fixture(scope="function")
    def schema_type(  # noqa: C901
        self, app, request, foo_schema, bar_schema, bar_optional_schema, bar_list_schema, bar_dict_schema
    ):
        if request.param is None:
            return None
        elif request.param == "bare_schema":
            return foo_schema.schema
        elif request.param == "schema":
            return t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema.schema)]
        elif request.param == "list_of_schema":
            return t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(foo_schema.schema)]
        elif request.param == "list_of_bare_schema":
            return list[foo_schema.schema]
        elif request.param == "schema_partial":
            if app.schema.schema_library.lib in (typesystem,):
                pytest.skip("Library does not support optional partial schemas")
            return t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema.schema, partial=True)]
        elif request.param == "schema_nested":
            return t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(bar_schema.schema)]
        elif request.param == "schema_nested_optional":
            if app.schema.schema_library.lib in (typesystem, marshmallow):
                pytest.skip("Library does not support optional nested schemas")
            return t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(bar_optional_schema.schema)]
        elif request.param == "schema_nested_list":
            return t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(bar_list_schema.schema)]
        elif request.param == "schema_nested_dict":
            return t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(bar_dict_schema.schema)]

        else:
            raise ValueError("Wrong schema type")

    @pytest.mark.parametrize(
        ["schema_type", "exception"],
        (
            pytest.param("bare_schema", None, id="bare_schema"),
            pytest.param("schema", None, id="schema"),
            pytest.param("schema_partial", None, id="schema_partial"),
            pytest.param("list_of_schema", None, id="list_of_schema"),
            pytest.param("list_of_bare_schema", None, id="list_of_bare_schema"),
            pytest.param("schema_nested", None, id="schema_nested"),
            pytest.param("schema_nested_optional", None, id="schema_nested_optional"),
            pytest.param("schema_nested_list", None, id="schema_nested_list"),
            pytest.param("schema_nested_dict", None, id="schema_nested_dict"),
            pytest.param(None, ValueError("Wrong schema type"), id="wrong"),
        ),
        indirect=["schema_type", "exception"],
    )
    def test_from_type(self, schema_type, exception):
        with exception:
            Schema.from_type(schema_type)

    def test_build(self, foo_schema):
        n = Mock()
        s = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            schemas_mock.adapter.build_schema.return_value = foo_schema.schema

            schema = Schema.build(n, s)

            assert schemas_mock.adapter.build_schema.call_args_list == [call(name=n, schema=s, fields={})]
            assert schema.schema == foo_schema.schema

    def test_is_schema(self):
        mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            Schema.is_schema(mock)

            assert schemas_mock.adapter.is_schema.call_args_list == [call(mock)]

    def test_name(self):
        mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            Schema(mock).name

            assert schemas_mock.adapter.name.call_args_list == [call(mock)]

    @pytest.mark.parametrize(
        ["schema_type", "json_schema", "key_to_replace"],
        (
            pytest.param(
                "schema",
                {"properties": {"name": {"type": "string"}}, "type": "object"},
                None,
                id="plain",
            ),
            pytest.param(
                "schema_partial",
                {"properties": {"name": {"anyOf": [{"type": "string"}, {"type": "null"}]}}, "type": "object"},
                None,
                id="partial",
            ),
            pytest.param(
                "list_of_schema",
                {"properties": {"name": {"type": "string"}}, "type": "object"},
                None,
                id="list",
            ),
            pytest.param(
                "list_of_bare_schema",
                {"properties": {"name": {"type": "string"}}, "type": "object"},
                None,
                id="list_bare",
            ),
            pytest.param(
                "schema_nested",
                {"properties": {"foo": {"$ref": "#/components/schemas/Foo"}}, "type": "object"},
                "properties.foo",
                id="nested",
            ),
            pytest.param(
                "schema_nested_optional",
                {
                    "properties": {"foo": {"anyOf": [{"$ref": "#/components/schemas/Foo"}, {"type": "null"}]}},
                    "type": "object",
                },
                "properties.foo.anyOf.0",
                id="nested_optional",
            ),
            pytest.param(
                "schema_nested_list",
                {
                    "properties": {"foo": {"items": {"$ref": "#/components/schemas/Foo"}, "type": "array"}},
                    "type": "object",
                },
                "properties.foo.items",
                id="nested_list",
            ),
            pytest.param(
                "schema_nested_dict",
                {
                    "properties": {
                        "foo": {"additionalProperties": {"$ref": "#/components/schemas/Foo"}, "type": "object"}
                    },
                    "type": "object",
                },
                "properties.foo.additionalProperties",
                id="nested_dict",
            ),
        ),
        indirect=["schema_type"],
    )
    def test_json_schema(self, schemas, schema_type, json_schema, key_to_replace):
        result = Schema.from_type(schema_type).json_schema({id(schemas["Foo"].schema): schemas["Foo"].name})

        expected_result = deepcopy(json_schema)

        if key_to_replace:
            subdict = functools.reduce(
                lambda x, k: x[int(k) if k.isnumeric() else k], key_to_replace.split("."), expected_result
            )
            subdict["$ref"] = subdict["$ref"].replace("Foo", schemas["Foo"].name)

        assert_recursive_contains(expected_result, result)

    def test_unique_schema(self):
        mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            Schema(mock).unique_schema

            assert schemas_mock.adapter.unique_schema.call_args_list == [call(mock)]

    def test_fields(self):
        mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            schemas_mock.adapter.unique_schema.return_value = mock

            Schema(mock).fields

            assert schemas_mock.adapter.unique_schema.call_args_list == [call(mock)]
            assert schemas_mock.adapter.schema_fields.call_args_list == [call(mock)]

    @pytest.mark.parametrize(
        ["schema", "nested_schemas"],
        (
            pytest.param("Foo", [], id="no_nested"),
            pytest.param("Bar", ["Foo"], id="attribute_nested"),
            pytest.param("BarList", ["Foo"], id="list_nested"),
            pytest.param("BarDict", ["Foo"], id="dict_nested"),
        ),
    )
    def test_nested_schemas(self, schemas, schema, nested_schemas):
        result = Schema(schemas[schema].schema).nested_schemas()

        assert result == [schemas[x].schema for x in nested_schemas]

    @pytest.mark.parametrize(
        ["values", "expected_result"],
        (
            pytest.param(Mock(partial=False), True, id="single"),
            pytest.param([Mock(), Mock()], [True, True], id="multiple"),
        ),
    )
    def test_validate(self, values, expected_result):
        schema_mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            schemas_mock.adapter.validate.return_value = True

            result = Schema(schema_mock).validate(values)

            assert result == expected_result
            assert schemas_mock.adapter.validate.call_args_list == [
                call(schema_mock, x, partial=False) for x in (values if isinstance(values, list) else [values])
            ]

    @pytest.mark.parametrize(
        ["values", "expected_result"],
        (
            pytest.param(Mock(), True, id="single"),
            pytest.param([Mock(), Mock()], [True, True], id="multiple"),
        ),
    )
    def test_load(self, values, expected_result):
        schema_mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            schemas_mock.adapter.load.return_value = True

            result = Schema(schema_mock).load(values)

            assert result == expected_result
            assert schemas_mock.adapter.load.call_args_list == [
                call(schema_mock, x) for x in (values if isinstance(values, list) else [values])
            ]

    @pytest.mark.parametrize(
        ["values", "expected_result"],
        (
            pytest.param(Mock(), True, id="single"),
            pytest.param([Mock(), Mock()], [True, True], id="multiple"),
        ),
    )
    def test_dump(self, values, expected_result):
        schema_mock = Mock()
        with patch("flama.schemas.data_structures.schemas") as schemas_mock:
            schemas_mock.adapter.dump.return_value = True

            result = Schema(schema_mock).dump(values)

            assert result == expected_result
            assert schemas_mock.adapter.dump.call_args_list == [
                call(schema_mock, x) for x in (values if isinstance(values, list) else [values])
            ]


class TestCaseParameter:
    @pytest.mark.parametrize(
        ["type_", "parameter", "result"],
        (
            pytest.param(
                "path",
                InjectionParameter("foo", str),
                Parameter(name="foo", location=ParameterLocation.path, type=str),
                id="path",
            ),
            pytest.param(
                "query",
                InjectionParameter("foo", int),
                Parameter(name="foo", location=ParameterLocation.query, type=int),
                id="query",
            ),
            pytest.param(
                "body",
                InjectionParameter("foo", Unknown),
                Parameter(name="foo", location=ParameterLocation.body, type=Unknown),
                id="body",
            ),
            pytest.param(
                "response",
                InjectionParameter("foo", Unknown),
                Parameter(name="foo", location=ParameterLocation.response, type=Unknown),
                id="response",
            ),
        ),
    )
    def test_build(self, foo_schema, type_, parameter, result):
        if parameter.annotation == Unknown:
            parameter = InjectionParameter(
                parameter.name,
                t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema.schema)],
                parameter.default,
            )
            result = Parameter(
                name=result.name,
                location=result.location,
                type=t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema.schema)],
                required=result.required,
                default=result.default,
            )

        assert Parameter.build(type_, parameter) == result



import datetime
import typing as t

import marshmallow
import marshmallow.validate
import pydantic
import pytest
import typesystem
import typesystem.fields

from flama import schemas
from tests.asserts import assert_recursive_contains

utc = datetime.timezone.utc


class TestCaseSchemaValidation:
    @pytest.fixture(scope="function")
    def product_schema(self, app):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model(
                "Product",
                name=(str, ...),
                rating=(t.Optional[int], None),
                created=(t.Optional[datetime.datetime], None),
            )
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(
                title="Product",
                fields={
                    "name": typesystem.fields.String(),
                    "rating": typesystem.fields.Integer(allow_null=True),
                    "created": typesystem.fields.DateTime(allow_null=True),
                },
            )
        elif app.schema.schema_library.lib == marshmallow:
            schema = type(
                "Product",
                (marshmallow.Schema,),
                {
                    "name": marshmallow.fields.String(),
                    "rating": marshmallow.fields.Integer(allow_none=True),
                    "created": marshmallow.fields.DateTime(allow_none=True),
                },
            )
        else:
            raise ValueError("Wrong schema lib")

        return schema

    @pytest.fixture(scope="function")
    def reviewed_product_schema(self, app, product_schema):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model("ReviewedProduct", reviewer=(str, ...), __base__=product_schema)
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(
                title="ReviewedProduct", fields={**product_schema.fields, **{"reviewer": typesystem.fields.String()}}
            )
        elif app.schema.schema_library.lib == marshmallow:
            schema = type("ReviewedProduct", (product_schema,), {"reviewer": marshmallow.fields.String()})
        else:
            raise ValueError("Wrong schema lib")

        return schema

    @pytest.fixture(scope="function")
    def location_schema(self, app):
        if app.schema.schema_library.lib == pydantic:

            def latitude_validator(cls, x):
                assert -90 <= x <= 90
                return x

            def longitude_validator(cls, x):
                assert -180 <= x <= 180
                return x

            schema = pydantic.create_model(
                "Location",
                latitude=(float, ...),
                longitude=(float, ...),
                __validators__={"latitude": latitude_validator, "longitude": longitude_validator},
            )
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(
                fields={
                    "latitude": typesystem.fields.Number(minimum=-90, maximum=90),
                    "longitude": typesystem.fields.Number(minimum=-180, maximum=180),
                }
            )
        elif app.schema.schema_library.lib == marshmallow:
            schema = type(
                "Location",
                (marshmallow.Schema,),
                {
                    "latitude": marshmallow.fields.Float(validate=marshmallow.validate.Range(min=-90, max=90)),
                    "longitude": marshmallow.fields.Float(validate=marshmallow.validate.Range(min=-180, max=180)),
                },
            )
        else:
            raise ValueError("Wrong schema lib")
        return schema

    @pytest.fixture(scope="function")
    def place_schema(self, app, location_schema):
        if app.schema.schema_library.lib == pydantic:
            schema = pydantic.create_model("Place", location=(location_schema, ...), name=(str, ...))
        elif app.schema.schema_library.lib == typesystem:
            schema = typesystem.Schema(
                fields={
                    "location": typesystem.Reference("Location", typesystem.Definitions({"Location": location_schema})),
                    "name": typesystem.String(),
                }
            )
        elif app.schema.schema_library.lib == marshmallow:
            schema = type(
                "Place",
                (marshmallow.Schema,),
                {"location": marshmallow.fields.Nested(location_schema), "name": marshmallow.fields.String()},
            )
        else:
            raise ValueError("Wrong schema lib")
        return schema

    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app, product_schema, reviewed_product_schema, place_schema):
        @app.route("/product", methods=["POST"])
        def product_identity(
            product: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(product_schema)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(product_schema)]:
            return product

        @app.route("/reviewed-product", methods=["POST"])
        def reviewed_product_identity(
            reviewed_product: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(reviewed_product_schema)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(reviewed_product_schema)]:
            return reviewed_product

        @app.route("/place", methods=["POST"])
        def place_identity(
            place: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(place_schema)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(place_schema)]:
            return place

        @app.route("/many-products", methods=["GET"])
        def many_products(
            products: t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(product_schema)],
        ) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(product_schema)]:
            return products

        @app.route("/partial-product", methods=["GET"])
        def partial_product(
            product: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(product_schema, partial=True)],
        ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(product_schema, partial=True)]:
            return product

        @app.route("/serialization-error")
        def serialization_error() -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(product_schema)]:
            return {"rating": "foo", "created": "bar"}

    @pytest.mark.parametrize(
        ["path", "method", "json_data", "expected_output", "status_code"],
        [
            pytest.param(
                "/product",
                "post",
                {
                    "name": "foo",
                    "rating": 0,
                    "created": "2018-01-01T00:00:00+00:00",
                },
                {
                    "name": "foo",
                    "rating": 0,
                    "created": "2018-01-01T00:00:00+00:00",
                },
                200,
                id="simple_schema",
            ),
            pytest.param(
                "/reviewed-product",
                "post",
                {
                    "name": "foo",
                    "rating": 0,
                    "created": "2018-01-01T00:00:00+00:00",
                    "reviewer": "bar",
                },
                {
                    "name": "foo",
                    "rating": 0,
                    "created": "2018-01-01T00:00:00+00:00",
                    "reviewer": "bar",
                },
                200,
                id="inherited_schema",
            ),
            pytest.param(
                "/place",
                "post",
                {"name": "foo", "location": {"latitude": 0.0, "longitude": 0.0}},
                {"name": "foo", "location": {"latitude": 0.0, "longitude": 0.0}},
                200,
                id="nested_schema",
            ),
            pytest.param(
                "/many-products",
                "get",
                [
                    {
                        "name": "foo",
                        "rating": 0,
                        "created": "2018-01-01T00:00:00+00:00",
                    },
                    {
                        "name": "bar",
                        "rating": 1,
                        "created": "2018-01-01T00:00:00+00:00",
                    },
                ],
                [
                    {
                        "name": "foo",
                        "rating": 0,
                        "created": "2018-01-01T00:00:00+00:00",
                    },
                    {
                        "name": "bar",
                        "rating": 1,
                        "created": "2018-01-01T00:00:00+00:00",
                    },
                ],
                200,
                id="array_schema",
            ),
            pytest.param(
                "/partial-product",
                "get",
                {"name": "foo"},
                {"name": "foo"},
                200,
                id="partial_schema",
            ),
            pytest.param(
                "/serialization-error",
                "get",
                None,
                {
                    "error": "SerializationError",
                    "status_code": 500,
                },
                500,
                id="serialization_error",
            ),
        ],
    )
    async def test_schemas(self, client, path, method, json_data, expected_output, status_code):
        response = await client.request(method, path, json=json_data)
        assert response.status_code == status_code, response.json()
        assert_recursive_contains(expected_output, response.json())



import functools
import inspect
import typing as t

from flama.injection.components import Component, Components
from flama.injection.exceptions import ComponentNotFound
from flama.injection.resolver import EMPTY, ROOT_NAME, Parameter, Resolver

if t.TYPE_CHECKING:
    from flama.injection.resolver import ResolutionTree


class Injector:
    """Functions dependency injector. It uses a resolver to generate dependencies trees and evaluate them."""

    def __init__(
        self,
        context_types: t.Optional[dict[str, type]] = None,
        components: t.Optional[t.Union[t.Sequence[Component], Components]] = None,
    ):
        """Functions dependency injector.

        It uses a resolver to generate dependencies trees and evaluate them.

        :param context_types: A mapping of names into types for injection contexts.
        :param components: List of components.
        """
        self.context_types = context_types or {}
        self.components = Components(components or [])
        self._resolver: t.Optional[Resolver] = None

    @property
    def context_types(self) -> dict[str, type]:
        return self._context_types

    @context_types.setter
    def context_types(self, context_types: dict[str, type]):
        self._context_types = context_types
        del self.resolver

    @context_types.deleter
    def context_types(self):
        self._context_types = {}
        del self.resolver

    @property
    def components(self) -> Components:
        return self._components

    @components.setter
    def components(self, components: Components):
        self._components = components
        del self.resolver

    @components.deleter
    def components(self):
        self._components = Components([])
        del self.resolver

    @property
    def resolver(self) -> Resolver:
        if self._resolver is None:
            self._resolver = Resolver(self.context_types, self.components)

        return self._resolver

    @resolver.deleter
    def resolver(self):
        self._resolver = None

    @t.overload
    def resolve(self, annotation: type) -> "ResolutionTree": ...
    @t.overload
    def resolve(self, annotation: type, *, name: str) -> "ResolutionTree": ...
    @t.overload
    def resolve(self, annotation: type, *, default: t.Any) -> "ResolutionTree": ...
    @t.overload
    def resolve(self, annotation: type, *, name: str, default: t.Any) -> "ResolutionTree": ...
    def resolve(self, annotation: type, *, name: t.Optional[str] = None, default: t.Any = EMPTY) -> "ResolutionTree":
        """Generate a dependencies tree for a given type annotation.

        :param annotation: Type annotation to be resolved.
        :param name: Name of the parameter to be resolved.
        :param func: Function to be resolved.
        :return: Dependencies tree.
        """
        return self.resolver.resolve(Parameter(name or ROOT_NAME, annotation, default))

    def resolve_function(self, func: t.Callable) -> dict[str, "ResolutionTree"]:
        """Generate a dependencies tree for a given function.

        It analyses the function signature, look for type annotations and try to resolve them.

        :param func: Function to be resolved.
        :return: Mapping of parameter names and dependencies trees.
        """
        parameters = {}
        for parameter in [
            x
            for x in inspect.signature(func).parameters.values()
            if not (x.name in ("self", "cls", "args", "kwargs") and x.annotation == inspect._empty)
        ]:
            try:
                parameters[parameter.name] = self.resolver.resolve(Parameter.from_parameter(parameter))
            except ComponentNotFound as e:
                raise ComponentNotFound(e.parameter, component=e.component, function=func) from None

        return parameters

    async def inject(self, func: t.Callable, context: t.Optional[dict[str, t.Any]] = None) -> t.Callable:
        """Inject dependencies into a given function.

        It analyses the function signature, look for type annotations and try to resolve them. Once all dependencies
        trees are resolved for every single parameter, it uses the given context to evaluate those trees and calculate
        a final value for each parameter. Finally, it returns a partialised function with all dependencies injected.

        :param func: Function to be partialised.
        :param context: Mapping of names and values used to gather injection values.
        :return: Partialised function with all dependencies injected.
        """
        if context is None:
            context = {}

        return functools.partial(
            func, **{name: await resolution.value(context) for name, resolution in self.resolve_function(func).items()}
        )



import asyncio
import dataclasses
import functools
import os
import ssl
import typing as t

import click
import uvicorn
from h11._connection import DEFAULT_MAX_INCOMPLETE_EVENT_SIZE
from uvicorn.config import (
    LOGGING_CONFIG,
    SSL_PROTOCOL_VERSION,
    HTTPProtocolType,
    InterfaceType,
    LifespanType,
    LoopSetupType,
    WSProtocolType,
)
from uvicorn.main import HTTP_CHOICES, INTERFACE_CHOICES, LEVEL_CHOICES, LIFESPAN_CHOICES, LOOP_CHOICES, WS_CHOICES

if t.TYPE_CHECKING:
    from flama.applications import Flama

__all__ = ["options", "Uvicorn"]

decorators = (
    click.option(
        "--server-host",
        type=str,
        default="127.0.0.1",
        envvar="HOST",
        help="Bind socket to this host.",
        show_default=True,
    ),
    click.option(
        "--server-port",
        type=int,
        default=8000,
        envvar="PORT",
        help="Bind socket to this port.",
        show_default=True,
    ),
    click.option("--server-reload", is_flag=True, default=False, help="Enable auto-reload."),
    click.option("--server-uds", type=str, default=None, help="Bind to a UNIX domain socket."),
    click.option("--server-fd", type=int, default=None, help="Bind to socket from this file descriptor."),
    click.option(
        "--server-reload-dirs",
        multiple=True,
        help="Set reload directories explicitly, instead of using the current working directory.",
        type=click.Path(exists=True),
    ),
    click.option(
        "--server-reload-includes",
        multiple=True,
        help="Set glob patterns to include while watching for files. Includes '*.py' "
        "by default; these defaults can be overridden with `--server-reload-exclude`. "
        "This option has no effect unless watchfiles is installed.",
    ),
    click.option(
        "--server-reload-excludes",
        multiple=True,
        help="Set glob patterns to exclude while watching for files. Includes "
        "'.*, .py[cod], .sw.*, ~*' by default; these defaults can be overridden "
        "with `--server-reload-include`. This option has no effect unless watchfiles is "
        "installed.",
    ),
    click.option(
        "--server-reload-delay",
        type=float,
        default=0.25,
        show_default=True,
        help="Delay between previous and next check if application needs to be. Defaults to 0.25s.",
    ),
    click.option(
        "--server-workers",
        default=None,
        type=int,
        help="Number of worker processes. Defaults to the $WEB_CONCURRENCY environment"
        " variable if available, or 1. Not valid with --server-dev.",
    ),
    click.option(
        "--server-loop",
        type=LOOP_CHOICES,
        default="auto",
        help="Event loop implementation.",
        show_default=True,
    ),
    click.option(
        "--server-http",
        type=HTTP_CHOICES,
        default="auto",
        help="HTTP protocol implementation.",
        show_default=True,
    ),
    click.option(
        "--server-ws",
        type=WS_CHOICES,
        default="auto",
        help="WebSocket protocol implementation.",
        show_default=True,
    ),
    click.option(
        "--server-ws-max-size",
        type=int,
        default=16777216,
        help="WebSocket max size message in bytes",
        show_default=True,
    ),
    click.option(
        "--server-ws-ping-interval",
        type=float,
        default=20.0,
        help="WebSocket ping interval",
        show_default=True,
    ),
    click.option(
        "--server-ws-ping-timeout",
        type=float,
        default=20.0,
        help="WebSocket ping timeout",
        show_default=True,
    ),
    click.option(
        "--server-ws-per-message-deflate",
        type=bool,
        default=True,
        help="WebSocket per-message-deflate compression",
        show_default=True,
    ),
    click.option(
        "--server-lifespan",
        type=LIFESPAN_CHOICES,
        default="auto",
        help="Lifespan implementation.",
        show_default=True,
    ),
    click.option(
        "--server-interface",
        type=INTERFACE_CHOICES,
        default="auto",
        help="Select ASGI3, ASGI2, or WSGI as the application interface.",
        show_default=True,
    ),
    click.option(
        "--server-env-file",
        type=click.Path(exists=True),
        default=None,
        help="Environment configuration file.",
        show_default=True,
    ),
    click.option(
        "--server-log-config",
        type=click.Path(exists=True),
        default=None,
        help="Logging configuration file. Supported formats: .ini, .json, .yaml.",
        show_default=True,
    ),
    click.option(
        "--server-log-level",
        type=LEVEL_CHOICES,
        default=None,
        help="Log level. [default: info]",
        show_default=True,
    ),
    click.option(
        "--server-access-log/--server-no-access-log",
        is_flag=True,
        default=True,
        help="Enable/Disable access log.",
    ),
    click.option(
        "--server-use-colors/--server-no-use-colors",
        is_flag=True,
        default=None,
        help="Enable/Disable colorized logging.",
    ),
    click.option(
        "--server-proxy-headers/--server-no-proxy-headers",
        is_flag=True,
        default=True,
        help="Enable/Disable X-Forwarded-Proto, X-Forwarded-For, X-Forwarded-Port to populate remote address info.",
    ),
    click.option(
        "--server-server-header/--server-no-server-header",
        is_flag=True,
        default=True,
        help="Enable/Disable default Server header.",
    ),
    click.option(
        "--server-date-header/--server-no-date-header",
        is_flag=True,
        default=True,
        help="Enable/Disable default Date header.",
    ),
    click.option(
        "--server-forwarded-allow-ips",
        type=str,
        default=None,
        help="Comma separated list of IPs to trust with proxy headers. Defaults to"
        " the $FORWARDED_ALLOW_IPS environment variable if available, or '127.0.0.1'.",
    ),
    click.option(
        "--server-root-path",
        type=str,
        default="",
        help="Set the ASGI 'root_path' for applications submounted below a given URL path.",
    ),
    click.option(
        "--server-limit-concurrency",
        type=int,
        default=None,
        help="Maximum number of concurrent connections or tasks to allow, before issuing HTTP 503 responses.",
    ),
    click.option(
        "--server-backlog",
        type=int,
        default=2048,
        help="Maximum number of connections to hold in backlog",
    ),
    click.option(
        "--server-limit-max-requests",
        type=int,
        default=None,
        help="Maximum number of requests to service before terminating the process.",
    ),
    click.option(
        "--server-timeout-keep-alive",
        type=int,
        default=5,
        help="Close Keep-Alive connections if no new data is received within this timeout.",
        show_default=True,
    ),
    click.option("--server-ssl-keyfile", type=str, default=None, help="SSL key file", show_default=True),
    click.option(
        "--server-ssl-certfile",
        type=str,
        default=None,
        help="SSL certificate file",
        show_default=True,
    ),
    click.option(
        "--server-ssl-keyfile-password",
        type=str,
        default=None,
        help="SSL keyfile password",
        show_default=True,
    ),
    click.option(
        "--server-ssl-version",
        type=int,
        default=int(SSL_PROTOCOL_VERSION),
        help="SSL version to use (see stdlib ssl module's)",
        show_default=True,
    ),
    click.option(
        "--server-ssl-cert-reqs",
        type=int,
        default=int(ssl.CERT_NONE),
        help="Whether client certificate is required (see stdlib ssl module's)",
        show_default=True,
    ),
    click.option(
        "--server-ssl-ca-certs",
        type=str,
        default=None,
        help="CA certificates file",
        show_default=True,
    ),
    click.option(
        "--server-ssl-ciphers",
        type=str,
        default="TLSv1",
        help="Ciphers to use (see stdlib ssl module's)",
        show_default=True,
    ),
    click.option(
        "--server-headers",
        multiple=True,
        help="Specify custom default HTTP response headers as a Name:Value pair",
    ),
    click.option(
        "--server-app-dir",
        default=".",
        show_default=True,
        help="Look for APP in the specified directory, by adding this to the PYTHONPATH."
        " Defaults to the current working directory.",
    ),
    click.option(
        "--server-h11-max-incomplete-event-size",
        type=int,
        default=DEFAULT_MAX_INCOMPLETE_EVENT_SIZE,
        help="For h11, the maximum number of bytes to buffer of an incomplete event.",
    ),
    click.option(
        "--server-factory",
        is_flag=True,
        default=False,
        help="Treat APP as an application factory, i.e. a () -> <ASGI app> callable.",
        show_default=True,
    ),
)


@dataclasses.dataclass
class Uvicorn:
    host: str = "127.0.0.1"
    port: int = 8000
    uds: t.Optional[str] = None
    fd: t.Optional[int] = None
    loop: LoopSetupType = "auto"
    http: t.Union[type[asyncio.Protocol], HTTPProtocolType] = "auto"
    ws: t.Union[type[asyncio.Protocol], WSProtocolType] = "auto"
    ws_max_size: int = 16777216
    ws_ping_interval: t.Optional[float] = 20.0
    ws_ping_timeout: t.Optional[float] = 20.0
    ws_per_message_deflate: bool = True
    lifespan: LifespanType = "auto"
    interface: InterfaceType = "auto"
    reload: bool = False
    reload_dirs: t.Optional[t.Union[list[str], str]] = None
    reload_includes: t.Optional[t.Union[list[str], str]] = None
    reload_excludes: t.Optional[t.Union[list[str], str]] = None
    reload_delay: float = 0.25
    workers: t.Optional[int] = None
    env_file: t.Optional[t.Union[str, os.PathLike]] = None
    log_config: t.Optional[t.Union[dict[str, t.Any], str]] = dataclasses.field(
        default_factory=lambda: LOGGING_CONFIG.copy()  # type: ignore[no-any-return]
    )
    log_level: t.Optional[t.Union[str, int]] = None
    access_log: bool = True
    proxy_headers: bool = True
    server_header: bool = True
    date_header: bool = True
    forwarded_allow_ips: t.Optional[str] = None
    root_path: str = ""
    limit_concurrency: t.Optional[int] = None
    backlog: int = 2048
    limit_max_requests: t.Optional[int] = None
    timeout_keep_alive: int = 5
    ssl_keyfile: t.Optional[str] = None
    ssl_certfile: t.Optional[t.Union[str, os.PathLike]] = None
    ssl_keyfile_password: t.Optional[str] = None
    ssl_version: int = SSL_PROTOCOL_VERSION
    ssl_cert_reqs: int = ssl.CERT_NONE
    ssl_ca_certs: t.Optional[str] = None
    ssl_ciphers: str = "TLSv1"
    headers: t.Optional[list[tuple[str, str]]] = None
    use_colors: t.Optional[bool] = None
    app_dir: t.Optional[str] = None
    factory: bool = False
    h11_max_incomplete_event_size: int = DEFAULT_MAX_INCOMPLETE_EVENT_SIZE

    def run(self, app: t.Union[str, "Flama"]):
        uvicorn.run(app, **dataclasses.asdict(self))


def options(command: t.Callable) -> t.Callable:
    """Decorate a click command with all uvicorn options.

    :param command: Command to be decorated.
    :return: Decorated command.
    """

    @functools.wraps(command)
    def _inner(
        server_host: str = "127.0.0.1",
        server_port: int = 8000,
        server_uds: t.Optional[str] = None,
        server_fd: t.Optional[int] = None,
        server_loop: LoopSetupType = "auto",
        server_http: t.Union[type[asyncio.Protocol], HTTPProtocolType] = "auto",
        server_ws: t.Union[type[asyncio.Protocol], WSProtocolType] = "auto",
        server_ws_max_size: int = 16777216,
        server_ws_ping_interval: t.Optional[float] = 20.0,
        server_ws_ping_timeout: t.Optional[float] = 20.0,
        server_ws_per_message_deflate: bool = True,
        server_lifespan: LifespanType = "auto",
        server_interface: InterfaceType = "auto",
        server_reload: bool = False,
        server_reload_dirs: t.Optional[t.Union[list[str], str]] = None,
        server_reload_includes: t.Optional[t.Union[list[str], str]] = None,
        server_reload_excludes: t.Optional[t.Union[list[str], str]] = None,
        server_reload_delay: float = 0.25,
        server_workers: t.Optional[int] = None,
        server_env_file: t.Optional[t.Union[str, os.PathLike]] = None,
        server_log_config: t.Optional[t.Union[dict[str, t.Any], str]] = None,
        server_log_level: t.Optional[t.Union[str, int]] = None,
        server_access_log: bool = True,
        server_proxy_headers: bool = True,
        server_server_header: bool = True,
        server_date_header: bool = True,
        server_forwarded_allow_ips: t.Optional[str] = None,
        server_root_path: str = "",
        server_limit_concurrency: t.Optional[int] = None,
        server_backlog: int = 2048,
        server_limit_max_requests: t.Optional[int] = None,
        server_timeout_keep_alive: int = 5,
        server_ssl_keyfile: t.Optional[str] = None,
        server_ssl_certfile: t.Optional[t.Union[str, os.PathLike]] = None,
        server_ssl_keyfile_password: t.Optional[str] = None,
        server_ssl_version: int = SSL_PROTOCOL_VERSION,
        server_ssl_cert_reqs: int = ssl.CERT_NONE,
        server_ssl_ca_certs: t.Optional[str] = None,
        server_ssl_ciphers: str = "TLSv1",
        server_headers: t.Optional[list[tuple[str, str]]] = None,
        server_use_colors: t.Optional[bool] = None,
        server_app_dir: t.Optional[str] = None,
        server_factory: bool = False,
        server_h11_max_incomplete_event_size: int = DEFAULT_MAX_INCOMPLETE_EVENT_SIZE,
        *args,
        **kwargs,
    ):
        command(
            uvicorn=Uvicorn(
                host=server_host,
                port=server_port,
                reload=server_reload,
                uds=server_uds,
                fd=server_fd,
                loop=server_loop,
                http=server_http,
                ws=server_ws,
                ws_max_size=server_ws_max_size,
                ws_ping_interval=server_ws_ping_interval,
                ws_ping_timeout=server_ws_ping_timeout,
                ws_per_message_deflate=server_ws_per_message_deflate,
                lifespan=server_lifespan,
                interface=server_interface,
                reload_dirs=server_reload_dirs,
                reload_includes=server_reload_includes,
                reload_excludes=server_reload_excludes,
                reload_delay=server_reload_delay,
                workers=server_workers,
                env_file=server_env_file,
                log_config=server_log_config if server_log_config is not None else LOGGING_CONFIG,
                log_level=server_log_level,
                access_log=server_access_log,
                proxy_headers=server_proxy_headers,
                server_header=server_server_header,
                date_header=server_date_header,
                forwarded_allow_ips=server_forwarded_allow_ips,
                root_path=server_root_path,
                limit_concurrency=server_limit_concurrency,
                backlog=server_backlog,
                limit_max_requests=server_limit_max_requests,
                timeout_keep_alive=server_timeout_keep_alive,
                ssl_keyfile=server_ssl_keyfile,
                ssl_certfile=server_ssl_certfile,
                ssl_keyfile_password=server_ssl_keyfile_password,
                ssl_version=server_ssl_version,
                ssl_cert_reqs=server_ssl_cert_reqs,
                ssl_ca_certs=server_ssl_ca_certs,
                ssl_ciphers=server_ssl_ciphers,
                headers=server_headers,
                use_colors=server_use_colors,
                app_dir=server_app_dir,
                h11_max_incomplete_event_size=server_h11_max_incomplete_event_size,
                factory=server_factory,
            ),
            *args,
            **kwargs,
        )

    return functools.reduce(lambda x, y: y(x), decorators[::-1], _inner)



from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import exceptions, types, url
from flama.applications import Flama
from flama.routing.router import Router
from flama.routing.routes.base import BaseRoute
from flama.routing.routes.http import Route
from flama.routing.routes.mount import Mount


class TestCaseMount:
    @pytest.fixture(scope="function")
    def app(self):
        return Flama(schema=None, docs=None)

    @pytest.fixture(scope="function")
    def mount(self, app):
        return Mount("/foo/{x:int}/", app, name="foo")

    @pytest.mark.parametrize(
        ["app", "routes", "exception"],
        (
            pytest.param(
                MagicMock(spec=Flama),
                None,
                None,
                id="app",
            ),
            pytest.param(
                None,
                [MagicMock(spec=Route)],
                None,
                id="routes",
            ),
            pytest.param(
                None,
                None,
                exceptions.ApplicationError("Either 'path' and 'app' or 'mount' variables are needed"),
                id="wrong",
            ),
        ),
        indirect=["exception"],
    )
    def test_init(self, app, routes, exception):
        with exception:
            mount = Mount("/foo/", app, routes=routes)

            if app is None and routes:
                app = Router(routes=routes)

            assert mount.app == app
            assert mount.path == "/foo/"

    @pytest.mark.parametrize(
        ["scope_type", "handle_call"],
        (
            pytest.param("http", True, id="http"),
            pytest.param("websocket", True, id="websocket"),
            pytest.param("lifespan", True, id="lifespan"),
            pytest.param("wrong", False, id="wrong"),
        ),
    )
    async def test_call(self, scope_type, handle_call, mount, asgi_scope, asgi_receive, asgi_send):
        scope = types.Scope({**asgi_scope, "type": scope_type})
        route_scope = types.Scope({"foo": "bar"})
        handle = AsyncMock()
        expected_calls = [call(types.Scope({**scope, **route_scope}), asgi_receive, asgi_send)] if handle_call else []

        with patch.object(mount, "handle", new=handle), patch.object(mount, "route_scope", return_value=route_scope):
            await mount(scope, asgi_receive, asgi_send)

        assert handle.call_args_list == expected_calls

    def test_eq(self, app):
        assert Mount("/", app, name="app_mock") == Mount("/", app, name="app_mock")
        assert Mount("/", app, name="app_mock") != Mount("/", app, name="bar")

    @pytest.mark.parametrize(
        ["app", "used"],
        (
            pytest.param(MagicMock(spec=Router), False, id="router"),
            pytest.param(MagicMock(spec=Flama, router=MagicMock(spec=Router, components=[])), True, id="app"),
        ),
    )
    def test_build(self, mount, app, used):
        root_app = MagicMock(spec=Flama)
        expected_calls = [call(app)] if used else [call(root_app)]

        route = MagicMock(spec=Route)
        mount.app = app
        mount.app.routes = [route]

        mount.build(root_app)

        assert route.build.call_args_list == expected_calls

    @pytest.mark.parametrize(
        ["scope_type", "path_match_return", "result"],
        (
            pytest.param("http", MagicMock(match=url.Path.Match.exact), BaseRoute.Match.full, id="match_exact-http"),
            pytest.param(
                "http", MagicMock(match=url.Path.Match.partial), BaseRoute.Match.full, id="match_partial-http"
            ),
            pytest.param("http", MagicMock(match=url.Path.Match.none), BaseRoute.Match.none, id="no_match-http"),
            pytest.param(
                "websocket", MagicMock(match=url.Path.Match.exact), BaseRoute.Match.full, id="match_exact-websocket"
            ),
            pytest.param(
                "websocket", MagicMock(match=url.Path.Match.partial), BaseRoute.Match.full, id="match_partial-websocket"
            ),
            pytest.param(
                "websocket", MagicMock(match=url.Path.Match.none), BaseRoute.Match.none, id="no_match-websocket"
            ),
            pytest.param("wrong", None, BaseRoute.Match.none, id="wrong_scope_type"),
        ),
    )
    def test_match(self, scope_type, path_match_return, result, asgi_scope, mount):
        asgi_scope["type"] = scope_type

        with patch.object(mount.path, "match", return_value=path_match_return):
            assert mount.match(asgi_scope) == result

    async def test_handle(self, mount, asgi_scope, asgi_receive, asgi_send):
        app = AsyncMock(spec=Flama)
        mount = Mount("/", app)

        await mount.handle(asgi_scope, asgi_receive, asgi_send)

        assert app.call_args_list == [call(asgi_scope, asgi_receive, asgi_send)]

    @pytest.mark.parametrize(
        ["app", "used"],
        (
            pytest.param(Router(), False, id="router"),
            pytest.param(Flama(docs=None, schema=None), True, id="app"),
        ),
    )
    def test_route_scope(self, mount, asgi_scope, app, used):
        def bar(): ...

        mount.app = app
        mount.app.add_route("/bar", bar)

        asgi_scope["path"] = "/foo/1/bar"
        route_scope = mount.route_scope(asgi_scope)

        assert route_scope == {
            "app": app if used else asgi_scope["app"],
            "path": "/bar",
            "root_path": "" if used else "/foo/1/",
        }

    @pytest.mark.parametrize(
        ["name", "params", "expected_url", "exception"],
        (
            pytest.param(
                "foo", {"x": 1, "path": "/foo"}, url.URL(scheme="http", path="/foo/1/"), None, id="match_full_name"
            ),
            pytest.param("foo:bar", {"x": 1}, url.URL(scheme="http", path="/foo/1/bar"), None, id="match_route"),
            pytest.param(
                "foo:nested",
                {"x": 1, "path": "/foo/"},
                url.URL(scheme="http", path="/foo/1/nested"),
                None,
                id="match_nested_app",
            ),
            pytest.param(
                "wrong",
                {"x": 1},
                None,
                exceptions.NotFoundException(params={"x": 1}, name="wrong"),
                id="not_found_mount",
            ),
            pytest.param(
                "foo:wrong",
                {"x": 1},
                None,
                exceptions.NotFoundException(params={"x": 1}, name="foo:wrong"),
                id="not_found_route",
            ),
        ),
        indirect=["exception"],
    )
    def test_resolve_url(self, name, params, expected_url, exception, mount):
        mount.app.add_route("/bar", MagicMock(), name="bar")
        mount.app.mount("/nested", MagicMock(), name="nested")

        with exception:
            assert mount.resolve_url(name, **params) == expected_url

    def test_routes(self, mount):
        route = MagicMock(spec=Route)

        mount.app = MagicMock(spec=Flama)
        mount.app.routes = [route]

        assert mount.routes == [route]



import abc
import enum
import functools
import logging
import typing as t

from flama import concurrency, exceptions, types, url
from flama.pagination import paginator
from flama.schemas.routing import RouteParametersMixin

if t.TYPE_CHECKING:
    from flama.applications import Flama

__all__ = ["BaseEndpointWrapper", "BaseRoute"]

logger = logging.getLogger(__name__)


class BaseEndpointWrapper(abc.ABC):
    def __init__(self, handler: types.Handler, *, pagination: t.Optional[types.Pagination] = None):
        """Wraps a function or endpoint into ASGI application.

        :param handler: Function or endpoint.
        :param pagination: Apply a pagination technique.
        """
        if pagination:
            handler = paginator.paginate(pagination, handler)

        self.handler = handler
        functools.update_wrapper(self, handler)

    def __get__(self, instance, owner):
        return functools.partial(self.__call__, instance)

    @abc.abstractmethod
    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        """
        ...

    def __eq__(self, other) -> bool:
        return isinstance(other, BaseEndpointWrapper) and self.handler == other.handler


class BaseRoute(abc.ABC, RouteParametersMixin):
    class Match(enum.Enum):
        none = enum.auto()
        partial = enum.auto()
        full = enum.auto()

    def __init__(
        self,
        path: t.Union[str, url.Path],
        app: types.App,
        *,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        tags: t.Optional[dict[str, t.Any]] = None,
    ):
        """A route definition of a http endpoint.

        :param path: URL path.
        :param app: ASGI application.
        :param name: Route name.
        :param include_in_schema: True if this route must be listed as part of the App schema.
        :param tags: Route tags.
        """
        self.path = url.Path(path)
        self.app = app
        self.endpoint = app.handler if isinstance(app, BaseEndpointWrapper) else app
        self.name = name
        self.include_in_schema = include_in_schema
        self.tags = tags or {}
        super().__init__()

    @abc.abstractmethod
    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None: ...

    def __eq__(self, other: t.Any) -> bool:
        return (
            isinstance(other, BaseRoute)
            and self.path == other.path
            and self.app == other.app
            and self.name == other.name
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path!r}, name={(self.name or '')!r})"

    def build(self, app: t.Optional["Flama"] = None) -> None:
        """Build step for routes.

        Just build the parameters' descriptor part of RouteParametersMixin.

        :param app: Flama app.
        """
        if app:
            self.parameters.build(app)

    def endpoint_handlers(self) -> dict[str, t.Callable]:
        """Return a mapping of all possible endpoints of this route.

        Useful to identify all endpoints by HTTP methods.

        :return: Mapping of all endpoints.
        """
        return {}

    async def handle(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request by calling the app of this route.

        :param scope: ASGI scope.
        :param receive: ASGI receive event.
        :param send: ASGI send event.
        """
        await concurrency.run(self.app, scope, receive, send)

    def match(self, scope: types.Scope) -> Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        return self.Match.full if self.path.match(scope["path"]).match == self.path.Match.exact else self.Match.none

    def route_scope(self, scope: types.Scope) -> types.Scope:
        """Build route scope from given scope.

        :param scope: ASGI scope.
        :return: Route scope.
        """
        return types.Scope({})

    def resolve_url(self, name: str, **params: t.Any) -> url.URL:
        """Builds URL path for given name and params.

        :param name: Route name.
        :param params: Path params.
        :return: URL path.
        """
        if name != self.name:
            raise exceptions.NotFoundException(params=params, name=name)

        try:
            result = self.path.build(**params)
        except ValueError:
            raise exceptions.NotFoundException(params=params, name=name)

        if result.unused:
            raise exceptions.NotFoundException(params=params, name=name)

        return url.URL(path=result.path, scheme="http")



import functools
import typing as t
from unittest.mock import MagicMock, call, patch

import pytest

from flama.injection.components import Component, Components
from flama.injection.exceptions import ComponentError, ComponentNotFound
from flama.injection.injector import Injector
from flama.injection.resolver import EMPTY, Parameter, ResolutionTree, Resolver

Foo = t.NewType("Foo", str)
Bar = t.NewType("Bar", str)
CustomStr = t.NewType("CustomStr", str)
Unknown = t.NewType("Unknown", str)


class LiteralComponent(Component):
    def resolve(self) -> Foo:
        return Foo("foo")


class ContextComponent(Component):
    def resolve(self, x: CustomStr) -> Foo:
        return Foo(x)


class ChildNestedComponent(Component):
    def resolve(self) -> Bar:
        return Bar("bar")


class NestedComponent(Component):
    def resolve(self, x: Bar) -> Foo:
        return Foo(x)


class UnhandledComponent(Component):
    def resolve(self):
        pass


def function(foo: Foo):
    return foo


class TestCaseInjector:
    @pytest.fixture(scope="function")
    def injector(self):
        return Injector()

    def test_context_types_property(self, injector):
        value = {"foo": str}
        assert injector._resolver is None
        injector.resolver
        assert injector._resolver is not None

        # Set a value
        injector.context_types = value
        assert injector._resolver is None
        injector.resolver
        assert injector._resolver is not None

        # Delete the value
        assert injector.context_types == value
        del injector.context_types
        assert injector._resolver is None
        injector.resolver
        assert injector._resolver is not None

    def test_components_property(self, injector):
        value = Components([LiteralComponent()])

        assert injector._resolver is None
        injector.resolver
        assert injector._resolver is not None

        # Set a value
        injector.components = value
        assert injector._resolver is None
        injector.resolver
        assert injector._resolver is not None

        # Delete the value
        assert injector.components == value
        del injector.components
        assert injector._resolver is None
        injector.resolver
        assert injector._resolver is not None

    def test_resolve(self):
        resolver_mock = MagicMock(spec=Resolver)
        resolution_mock = MagicMock(spec=ResolutionTree)
        resolver_mock.resolve.return_value = resolution_mock

        injector = Injector()
        with patch.object(injector, "_resolver", resolver_mock):
            resolution = injector.resolve(Foo, name="foo", default=Foo("foo"))

        assert resolver_mock.resolve.call_args_list == [call(Parameter("foo", Foo, Foo("foo")))]
        assert resolution == resolution_mock

    def test_resolve_function(self):
        resolver_mock = MagicMock(spec=Resolver)
        resolution_mock = MagicMock(spec=ResolutionTree)
        resolver_mock.resolve.return_value = resolution_mock

        injector = Injector()
        with patch.object(injector, "_resolver", resolver_mock):
            resolution = injector.resolve_function(function)

        assert resolver_mock.resolve.call_args_list == [call(Parameter("foo", Foo, EMPTY))]
        assert resolution == {"foo": resolution_mock}

    @pytest.mark.parametrize(
        ["context", "context_types", "components", "result", "exception"],
        (
            pytest.param({"foo": Foo("foo")}, {"foo": Foo}, Components(), Foo("foo"), None, id="context"),
            pytest.param({}, {}, Components([LiteralComponent()]), Foo("foo"), None, id="component"),
            pytest.param(
                {"x": CustomStr("bar")},
                {"x": CustomStr},
                Components([ContextComponent()]),
                Foo("bar"),
                None,
                id="component_context",
            ),
            pytest.param(
                {}, {}, Components([NestedComponent(), ChildNestedComponent()]), Foo("bar"), None, id="component_nested"
            ),
            pytest.param(
                {},
                {},
                Components([UnhandledComponent()]),
                None,
                ComponentError(
                    "Component 'UnhandledComponent' must include a return annotation on the 'resolve' method, or "
                    "override 'can_handle_parameter'"
                ),
                id="unhandled",
            ),
            pytest.param(
                {},
                {},
                Components(),
                None,
                ComponentNotFound(Parameter("foo"), function=function),
                id="unknown_parameter",
            ),
            pytest.param(
                {},
                {},
                Components([NestedComponent()]),
                None,
                (
                    ComponentNotFound,
                    "No component able to handle parameter 'x' in component 'NestedComponent' for function 'function'",
                ),
                id="unknown_parameter_in_component",
            ),
        ),
        indirect=["exception"],
    )
    async def test_inject(self, context, context_types, components, result, exception):
        with exception:
            injector = Injector(context_types, components)
            injected_func = await injector.inject(function, context)

            assert isinstance(injected_func, functools.partial)
            assert injected_func.func == function
            assert "foo" in injected_func.keywords
            assert injected_func() == result



import collections
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import exceptions, types, url
from flama.applications import Flama
from flama.client import Client
from flama.endpoints import HTTPEndpoint, WebSocketEndpoint
from flama.injection import Component, Components
from flama.lifespan import Lifespan
from flama.routing.router import Router
from flama.routing.routes.http import Route
from flama.routing.routes.mount import Mount
from flama.routing.routes.websocket import WebSocketRoute


class TestCaseRouter:
    @pytest.fixture(scope="function")
    def app(self):
        return Flama(schema=None, docs=None)

    @pytest.fixture(scope="function")
    def router(self, app):
        return Router(root=app)

    @pytest.fixture(scope="function")
    def app_mock(self):
        return MagicMock(spec=Flama, router=MagicMock(spec=Router, components=Components([])))

    @pytest.fixture(scope="function")
    def component_mock(self):
        return MagicMock(spec=Component)

    @pytest.fixture(scope="function")
    def tags(self):
        return {"tag": "foo", "list_tag": ["foo", "bar"], "dict_tag": {"foo": "bar"}}

    @pytest.fixture(scope="function")
    def endpoint(self, request, router):  # noqa: C901
        if request.param == "function":

            async def foo():
                return "foo"

            def mount(app):
                app.add_route("/foo/", foo)

            endpoint = foo
        elif request.param == "endpoint":

            class FooHTTPEndpoint(HTTPEndpoint):
                async def get(self):
                    return "foo"

            def mount(app):
                app.add_route("/foo/", FooHTTPEndpoint)

            endpoint = FooHTTPEndpoint
        elif request.param == "websocket_function":

            async def foo():
                return "foo"

            def mount(app):
                app.add_websocket_route("/foo/", foo)

            endpoint = foo
        elif request.param == "websocket_endpoint":

            class FooWebSocketEndpoint(WebSocketEndpoint): ...

            def mount(app):
                app.add_websocket_route("/foo/", FooWebSocketEndpoint)

            endpoint = FooWebSocketEndpoint
        elif request.param == "nested_app":

            async def foo():
                return "foo"

            nested = Flama()
            nested.add_route("/foo/", foo)

            def mount(app):
                app.mount("/nested", app=nested)

            endpoint = foo
        elif request.param == "nested_router":

            async def foo():
                return "foo"

            router.add_route("/foo/", foo)

            def mount(app):
                app.mount("/router", app=router)

            endpoint = foo
        elif request.param == "nested_app_nested_router":

            async def foo():
                return "foo"

            nested = Flama()
            router.add_route("/foo/", foo)
            nested.mount("/router", app=router)

            def mount(app):
                app.mount("/nested", app=nested)

            endpoint = foo
        elif request.param == "wrong":

            class Foo: ...

            def mount(app):
                app.add_route("/wrong/", Foo)

            endpoint = Foo
        else:
            raise ValueError(f"Wrong value: {request.param}")

        return collections.namedtuple("Endpoint", ("endpoint", "mount"))(endpoint, mount)

    def test_init(self, app_mock):
        with patch("flama.routing.Router.build") as method_mock:
            Router([], root=app_mock)

        assert method_mock.call_args_list == [call(app_mock)]

    def test_eq(self):
        route = MagicMock(Route)
        assert Router(routes=[route]) == Router(routes=[route])

    @pytest.mark.parametrize(
        ["request_type"],
        (
            pytest.param("http", id="http"),
            pytest.param("websocket", id="websocket"),
        ),
    )
    async def test_call_route(self, request_type, router, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = request_type

        route = AsyncMock()
        route_scope = types.Scope({})
        with patch.object(router, "resolve_route", return_value=(route, route_scope)):
            await router(asgi_scope, asgi_receive, asgi_send)

        assert route.call_args_list == [call(route_scope, asgi_receive, asgi_send)]

    async def test_call_lifespan(self, router, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "lifespan"

        lifespan_mock = AsyncMock(Lifespan)
        with patch.object(router, "lifespan", lifespan_mock):
            await router(asgi_scope, asgi_receive, asgi_send)

        assert lifespan_mock.await_args_list == [call(asgi_scope, asgi_receive, asgi_send)]

    def test_build(self, router):
        app = MagicMock(Flama)
        route = MagicMock(Route)

        # First call to build when adding route
        router.add_route(route=route, root=app)

        # Second call to build
        router.build(app)

        assert route.build.call_args_list == [call(app), call(app)]

    def test_components(self, router, component_mock):
        assert router.components == []

        router.add_component(component_mock)
        leaf_component = MagicMock(spec=Component)
        leaf_router = Router(components=[leaf_component])
        router.mount("/app/", app=leaf_router)

        assert router.components == [component_mock]
        # Components are not propagated to leaf because mounted was done through a router instead of app
        assert leaf_router.components == [leaf_component]

    def test_add_component(self, router, component_mock):
        assert router.components == []

        router.add_component(component_mock)

        assert router.components == [component_mock]

    @pytest.mark.parametrize(
        ["endpoint", "path", "exception"],
        [
            pytest.param(
                "function",
                "/",
                None,
                id="function",
            ),
            pytest.param(
                "endpoint",
                "/",
                None,
                id="endpoint",
            ),
            pytest.param(
                "function",
                None,
                exceptions.ApplicationError("Either 'path' and 'endpoint' or 'route' variables are needed"),
                id="no_path",
            ),
            pytest.param(
                "wrong",
                "/",
                exceptions.ApplicationError("Endpoint must be a callable or an HTTPEndpoint subclass"),
                id="wrong_endpoint",
            ),
        ],
        indirect=["endpoint", "exception"],
    )
    def test_add_route(self, router, tags, endpoint, path, exception):
        with exception:
            router.add_route(path, endpoint.endpoint, tags=tags)

            assert len(router.routes) == 1
            assert isinstance(router.routes[0], Route)
            assert router.routes[0].path == path
            assert router.routes[0].endpoint == endpoint.endpoint
            assert router.routes[0].tags == tags

    @pytest.mark.parametrize(
        ["endpoint", "path", "exception"],
        [
            pytest.param(
                "function",
                "/",
                None,
                id="function",
            ),
            pytest.param(
                "endpoint",
                "/",
                None,
                id="endpoint",
            ),
            pytest.param(
                "function",
                None,
                exceptions.ApplicationError("Either 'path' and 'endpoint' or 'route' variables are needed"),
                id="no_path",
            ),
            pytest.param(
                "wrong",
                "/",
                exceptions.ApplicationError("Endpoint must be a callable or an HTTPEndpoint subclass"),
                id="wrong_endpoint",
            ),
        ],
        indirect=["endpoint", "exception"],
    )
    def test_route(self, router, tags, endpoint, path, exception):
        with exception:
            router.route(path, tags=tags)(endpoint.endpoint)

            assert len(router.routes) == 1
            assert isinstance(router.routes[0], Route)
            assert router.routes[0].path == path
            assert router.routes[0].endpoint == endpoint.endpoint
            assert router.routes[0].tags == tags

    @pytest.mark.parametrize(
        ["endpoint", "path", "exception"],
        [
            pytest.param(
                "websocket_function",
                "/",
                None,
                id="function",
            ),
            pytest.param(
                "websocket_endpoint",
                "/",
                None,
                id="endpoint",
            ),
            pytest.param(
                "function",
                None,
                exceptions.ApplicationError("Either 'path' and 'endpoint' or 'route' variables are needed"),
                id="no_path",
            ),
            pytest.param(
                "wrong",
                "/",
                exceptions.ApplicationError("Endpoint must be a callable or a WebSocketEndpoint subclass"),
                id="wrong_endpoint",
            ),
        ],
        indirect=["endpoint", "exception"],
    )
    def test_add_websocket(self, router, tags, endpoint, path, exception):
        with exception:
            router.add_websocket_route(path, endpoint.endpoint, tags=tags)

            assert len(router.routes) == 1
            assert isinstance(router.routes[0], WebSocketRoute)
            assert router.routes[0].path == path
            assert router.routes[0].endpoint == endpoint.endpoint
            assert router.routes[0].tags == tags

    @pytest.mark.parametrize(
        ["endpoint", "path", "exception"],
        [
            pytest.param(
                "websocket_function",
                "/",
                None,
                id="function",
            ),
            pytest.param(
                "websocket_endpoint",
                "/",
                None,
                id="endpoint",
            ),
            pytest.param(
                "function",
                None,
                exceptions.ApplicationError("Either 'path' and 'endpoint' or 'route' variables are needed"),
                id="no_path",
            ),
            pytest.param(
                "wrong",
                "/",
                exceptions.ApplicationError("Endpoint must be a callable or a WebSocketEndpoint subclass"),
                id="wrong_endpoint",
            ),
        ],
        indirect=["endpoint", "exception"],
    )
    def test_websocket_route_function(self, router, tags, endpoint, path, exception):
        with exception:
            router.websocket_route(path, tags=tags)(endpoint.endpoint)

            assert len(router.routes) == 1
            assert isinstance(router.routes[0], WebSocketRoute)
            assert router.routes[0].path == path
            assert router.routes[0].endpoint == endpoint.endpoint
            assert router.routes[0].tags == tags

    def test_mount_app(self, app, app_mock, tags):
        app.mount("/app/", app=app_mock, tags=tags)

        assert len(app.routes) == 1
        assert isinstance(app.routes[0], Mount)
        assert app.routes[0].path == "/app/"
        assert app.routes[0].app == app_mock
        assert app.routes[0].tags == tags

    def test_mount_router(self, app, component_mock, tags):
        router = Router(components=[component_mock])

        app.mount("/app/", app=router, tags=tags)

        assert len(app.router.routes) == 1
        # Check mount is initialized
        assert isinstance(app.routes[0], Mount)
        mount_route = app.router.routes[0]
        assert mount_route.path == "/app/"
        assert mount_route.tags == tags
        # Check router is created and initialized, also shares components and modules with main app
        assert isinstance(mount_route.app, Router)
        mount_router = mount_route.app
        default_components = app.components[:1]
        assert mount_router.components == [component_mock]
        assert app.components == default_components

    def test_mount_declarative(self, component_mock, tags):
        def root(): ...

        def foo(): ...

        def foo_view(): ...

        routes = [
            Route("/", root, tags=tags),
            Mount(
                "/foo",
                routes=[Route("/", foo, methods=["GET"]), Route("/view", foo_view, methods=["GET"])],
                components=[component_mock],
                tags=tags,
            ),
            Mount(
                "/bar",
                app=Router(
                    routes=[Route("/", foo, methods=["GET"]), Route("/view", foo_view, methods=["GET"])],
                    components=[component_mock],
                ),
                tags=tags,
            ),
        ]

        app = Flama(routes=routes, schema=None, docs=None)

        assert len(app.router.routes) == 3

        # Check first-level route is initialized
        assert isinstance(app.router.routes[0], Route)
        root_route = app.router.routes[0]
        assert root_route.path == "/"
        assert root_route.tags == tags

        # Check mount with routes is initialized
        assert isinstance(app.router.routes[1], Mount)
        mount_with_routes_route = app.router.routes[1]
        assert mount_with_routes_route.path == "/foo"
        assert mount_with_routes_route.tags == tags
        # Check router is created and initialized, also shares components and modules with main app
        assert isinstance(mount_with_routes_route.app, Router)
        mount_with_routes_router = mount_with_routes_route.app
        assert mount_with_routes_router.components == [component_mock]
        default_components = app.components[:1]
        assert app.components == default_components
        # Check second-level routes are created an initialized
        assert len(mount_with_routes_route.routes) == 2
        assert mount_with_routes_route.routes[0].path == "/"
        assert mount_with_routes_route.routes[1].path == "/view"

        # Check mount with app is initialized
        assert isinstance(app.router.routes[2], Mount)
        mount_with_app_route = app.router.routes[2]
        assert mount_with_app_route.path == "/bar"
        assert mount_with_app_route.tags == tags
        # Check router is created and initialized, also shares components and modules with main app
        assert isinstance(mount_with_app_route.app, Router)
        mount_with_app_router = mount_with_app_route.app
        assert mount_with_app_router.components == [component_mock]
        assert app.components == default_components
        # Check second-level routes are created an initialized
        assert len(mount_with_app_route.routes) == 2
        assert mount_with_app_route.routes[0].path == "/"
        assert mount_with_app_route.routes[1].path == "/view"

    @pytest.mark.parametrize(
        ["endpoint", "path", "root_path", "endpoint_path", "method", "exception"],
        [
            pytest.param("function", "/foo/", "", "/foo/", "GET", None, id="function"),
            pytest.param("endpoint", "/foo/", "", "/foo/", "GET", None, id="endpoint"),
            pytest.param("nested_app", "/nested/foo/", "", "/foo/", "GET", None, id="nested_app"),
            pytest.param("nested_router", "/router/foo/", "/router", "/foo/", "GET", None, id="nested_router"),
            pytest.param(
                "nested_app_nested_router",
                "/nested/router/foo/",
                "/router",
                "/foo/",
                "GET",
                None,
                id="nested_app_nested_router",
            ),
            pytest.param("function", "/foo/", "", "/foo/", "POST", exceptions.MethodNotAllowedException, id="partial"),
            pytest.param("function", "/bar/", "", "/foo/", "GET", exceptions.NotFoundException, id="not_found"),
        ],
        indirect=["endpoint", "exception"],
    )
    def test_resolve_route(self, app, endpoint, path, root_path, endpoint_path, method, asgi_scope, exception):
        endpoint.mount(app)

        asgi_scope["path"] = path
        asgi_scope["method"] = method

        with exception:
            route, route_scope = app.router.resolve_route(scope=asgi_scope)

            assert route.endpoint == endpoint.endpoint
            assert route.path == "/foo/"
            assert route_scope["root_path"] == root_path
            assert route_scope["path"] == endpoint_path

    @pytest.mark.parametrize(
        ["routes", "result", "exception"],
        (
            pytest.param(
                [
                    MagicMock(spec=Route, resolve_url=MagicMock(return_value=url.URL("/foo"))),
                    MagicMock(spec=Route, resolve_url=MagicMock(side_effect=exceptions.NotFoundException)),
                ],
                url.URL("/foo"),
                None,
                id="first",
            ),
            pytest.param(
                [
                    MagicMock(spec=Route, resolve_url=MagicMock(side_effect=exceptions.NotFoundException)),
                    MagicMock(spec=Route, resolve_url=MagicMock(return_value=url.URL("/foo"))),
                ],
                url.URL("/foo"),
                None,
                id="other",
            ),
            pytest.param(
                [
                    MagicMock(spec=Route, resolve_url=MagicMock(side_effect=exceptions.NotFoundException)),
                    MagicMock(spec=Route, resolve_url=MagicMock(side_effect=exceptions.NotFoundException)),
                ],
                None,
                exceptions.NotFoundException,
                id="not_found",
            ),
        ),
        indirect=["exception"],
    )
    def test_resolve_url(self, router, routes, result, exception):
        for route in routes:
            router.add_route(route=route)

        with exception:
            assert router.resolve_url("foo") == result

    async def test_request_nested_app(self, app):
        sub_app = Flama(docs=None, schema=None)
        sub_app.add_route("/bar/", lambda: {"foo": "bar"}, ["GET"])

        app.mount("/foo", sub_app)

        async with Client(app) as client:
            response = await client.get("/foo/bar/")
            assert response.status_code == 200
            assert response.json() == {"foo": "bar"}



import typing as t

import starlette.websockets

from flama import exceptions, types, websockets
from flama.endpoints.base import BaseEndpoint

__all__ = ["WebSocketEndpoint"]


class WebSocketEndpoint(BaseEndpoint, types.WebSocketEndpointProtocol):
    encoding: t.Optional[types.Encoding] = None

    def __init__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:
        """A websocket endpoint.

        :param scope: ASGI scope.
        :param receive: ASGI receive function.
        :param send: ASGI send function.
        """
        if scope["type"] != "websocket":
            raise ValueError("Wrong scope")

        super().__init__(scope, receive, send)

        self.state.update(
            {
                "websocket": websockets.WebSocket(self.state["scope"], receive, send),
                "websocket_encoding": self.encoding,
                "websocket_code": None,
                "websocket_message": None,
            }
        )

    @classmethod
    def allowed_handlers(cls) -> dict[str, t.Callable]:
        """A mapping of handler related to each WS action.

        :return: Handlers mapping.
        """
        return {
            "WEBSOCKET_CONNECT": cls.on_connect,
            "WEBSOCKET_RECEIVE": cls.on_receive,
            "WEBSOCKET_DISCONNECT": cls.on_disconnect,
        }

    async def dispatch(self) -> None:
        """Dispatch a request."""
        app = self.state["app"]
        websocket = self.state["websocket"]

        on_connect = await app.injector.inject(self.on_connect, self.state)
        await on_connect()

        try:
            self.state["websocket_message"] = await websocket.receive()

            while websocket.is_connected:
                on_receive = await app.injector.inject(self.on_receive, self.state)
                await on_receive()
                self.state["websocket_message"] = await websocket.receive()

            self.state["websocket_code"] = types.Code(int(self.state["websocket_message"].get("code", 1000)))
        except starlette.websockets.WebSocketDisconnect as e:
            self.state["websocket_code"] = types.Code(e.code)
            raise exceptions.WebSocketException(e.code, e.reason) from None
        except exceptions.WebSocketException as e:
            self.state["websocket_code"] = types.Code(e.code)
            raise e from None
        except Exception as e:
            self.state["websocket_code"] = types.Code(1011)
            raise e from None
        finally:
            on_disconnect = await app.injector.inject(self.on_disconnect, self.state)
            await on_disconnect()

    async def on_connect(self, websocket: websockets.WebSocket, *args, **kwargs) -> None:
        """Handle an incoming websocket connection.

        :param websocket: Websocket.
        """
        await websocket.accept()

    async def on_receive(self, websocket: websockets.WebSocket, data: types.Data, *args, **kwargs) -> None:
        """Handle an incoming websocket message.

        :param websocket: Websocket.
        :param data: Received data.
        """
        ...

    async def on_disconnect(self, websocket: websockets.WebSocket, websocket_code: types.Code, *args, **kwargs) -> None:
        """Handle a disconnecting websocket.

        :param websocket: Websocket.
        :param websocket_code: Websocket closing code.
        """
        await websocket.close(websocket_code)



import pytest

from flama.models.components import ModelComponentBuilder
from flama.models.models.pytorch import PyTorchModel
from flama.models.models.sklearn import SKLearnModel
from flama.models.models.tensorflow import TensorFlowModel


class TestCaseModelComponent:
    @pytest.mark.parametrize(
        ("model_path", "component_model_class", "serialized_model_class"),
        (
            pytest.param("sklearn", SKLearnModel, "sklearn", id="sklearn"),
            pytest.param("sklearn-pipeline", SKLearnModel, "sklearn-pipeline", id="sklearn-pipeline"),
            pytest.param("tensorflow", TensorFlowModel, "tensorflow", id="tensorflow"),
            pytest.param("torch", PyTorchModel, "torch", id="torch"),
        ),
        indirect=["model_path", "serialized_model_class"],
    )
    def test_build(self, model_path, component_model_class, serialized_model_class):
        component = ModelComponentBuilder.load(model_path)

        model = component.model
        model_instance = model.model
        assert isinstance(model, component_model_class)
        assert isinstance(model_instance, serialized_model_class)



import typing as t

import marshmallow
import pydantic
import pytest
import typesystem
import typesystem.fields

import flama.types.websockets
from flama import Component, endpoints, routing, schemas, websockets
from flama.schemas.data_structures import Parameter, ParameterLocation


class Custom: ...


class TestCaseRouteFieldsMixin:
    @pytest.fixture
    def component(self):
        class CustomComponent(Component):
            def resolve(self, ax: int) -> Custom:
                return Custom()

        return CustomComponent()

    @pytest.fixture
    def foo_schema(self, app):
        from flama import schemas

        if schemas.lib == pydantic:
            schema = pydantic.create_model("FooSchema", x=(int, ...), y=(str, ...))
        elif schemas.lib == typesystem:
            schema = typesystem.Schema(
                title="FooSchema", fields={"x": typesystem.fields.Integer(), "y": typesystem.fields.String()}
            )
        elif schemas.lib == marshmallow:
            schema = type(
                "FooSchema",
                (marshmallow.Schema,),
                {"x": marshmallow.fields.Integer(), "y": marshmallow.fields.String()},
            )
        else:
            raise ValueError("Wrong schema lib")

        return schema

    @pytest.fixture
    def route(self, request, foo_schema):
        if request.param == "http_function":

            def foo(
                w: int,
                a: Custom,
                z: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)],
                x: int = 1,
                y: t.Optional[str] = None,
            ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)]: ...

            return routing.Route("/foo/{w:int}/", endpoint=foo, methods=["GET"])

        elif request.param == "http_endpoint":

            class FooEndpoint(endpoints.HTTPEndpoint):
                def get(
                    self,
                    w: int,
                    a: Custom,
                    z: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)],
                    x: int = 1,
                    y: t.Optional[str] = None,
                ) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)]: ...

            return routing.Route("/bar/{w:int}/", endpoint=FooEndpoint, methods=["GET"])

        elif request.param == "websocket_function":

            def bar(
                websocket: websockets.WebSocket,
                data: flama.types.websockets.Data,
                w: int,
                a: Custom,
                z: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)],
                x: int = 1,
                y: t.Optional[str] = None,
            ) -> None: ...

            return routing.WebSocketRoute("/foo/{w:int}/", endpoint=bar)

        elif request.param == "websocket_endpoint":

            class FooWebsocket(endpoints.WebSocketEndpoint):
                async def on_receive(
                    self,
                    websocket: websockets.WebSocket,
                    data: flama.types.websockets.Data,
                    w: int,
                    a: Custom,
                    z: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)],
                    x: int = 1,
                    y: t.Optional[str] = None,
                ) -> None: ...

            return routing.WebSocketRoute("/foo/{w:int}/", endpoint=FooWebsocket)
        else:
            raise ValueError("Wrong value")

    @pytest.fixture(scope="function", autouse=True)
    def add_component(self, app, component):
        app.add_component(component)

    @pytest.mark.parametrize(
        ["route", "expected_params"],
        (
            pytest.param(
                "http_function",
                {
                    "GET": {
                        "x": {
                            "name": "x",
                            "location": ParameterLocation.query,
                            "type": int,
                            "required": False,
                            "default": 1,
                        },
                        "y": {
                            "name": "y",
                            "location": ParameterLocation.query,
                            "type": t.Optional[str],
                            "required": False,
                            "default": None,
                        },
                        "ax": {"name": "ax", "location": ParameterLocation.query, "type": int, "required": True},
                    },
                    "HEAD": {
                        "x": {
                            "name": "x",
                            "location": ParameterLocation.query,
                            "type": int,
                            "required": False,
                            "default": 1,
                        },
                        "y": {
                            "name": "y",
                            "location": ParameterLocation.query,
                            "type": t.Optional[str],
                            "required": False,
                            "default": None,
                        },
                        "ax": {"name": "ax", "location": ParameterLocation.query, "type": int, "required": True},
                    },
                },
                id="http_function",
            ),
            pytest.param(
                "http_endpoint",
                {
                    "GET": {
                        "x": {
                            "name": "x",
                            "location": ParameterLocation.query,
                            "type": int,
                            "required": False,
                            "default": 1,
                        },
                        "y": {
                            "name": "y",
                            "location": ParameterLocation.query,
                            "type": t.Optional[str],
                            "required": False,
                            "default": None,
                        },
                        "ax": {"name": "ax", "location": ParameterLocation.query, "type": int, "required": True},
                    },
                    "HEAD": {
                        "x": {
                            "name": "x",
                            "location": ParameterLocation.query,
                            "type": int,
                            "required": False,
                            "default": 1,
                        },
                        "y": {
                            "name": "y",
                            "location": ParameterLocation.query,
                            "type": t.Optional[str],
                            "required": False,
                            "default": None,
                        },
                        "ax": {"name": "ax", "location": ParameterLocation.query, "type": int, "required": True},
                    },
                },
                id="http_endpoint",
            ),
            pytest.param(
                "websocket_function",
                {
                    "WEBSOCKET": {
                        "x": {
                            "name": "x",
                            "location": ParameterLocation.query,
                            "type": int,
                            "required": False,
                            "default": 1,
                        },
                        "y": {
                            "name": "y",
                            "location": ParameterLocation.query,
                            "type": t.Optional[str],
                            "required": False,
                            "default": None,
                        },
                        "ax": {"name": "ax", "location": ParameterLocation.query, "type": int, "required": True},
                    }
                },
                id="websocket_function",
            ),
            pytest.param(
                "websocket_endpoint",
                {
                    "WEBSOCKET_CONNECT": {},
                    "WEBSOCKET_RECEIVE": {
                        "x": {
                            "name": "x",
                            "location": ParameterLocation.query,
                            "type": int,
                            "required": False,
                            "default": 1,
                        },
                        "y": {
                            "name": "y",
                            "location": ParameterLocation.query,
                            "type": t.Optional[str],
                            "required": False,
                            "default": None,
                        },
                        "ax": {"name": "ax", "location": ParameterLocation.query, "type": int, "required": True},
                    },
                    "WEBSOCKET_DISCONNECT": {},
                },
                id="websocket_endpoint",
            ),
        ),
        indirect=["route"],
    )
    def test_query(self, route, expected_params):
        assert route.parameters.query == {
            method: {k: Parameter(**param) for k, param in params.items()} for method, params in expected_params.items()
        }

    @pytest.mark.parametrize(
        ["route", "expected_params"],
        (
            pytest.param(
                "http_function",
                {
                    "GET": {
                        "w": {"name": "w", "location": ParameterLocation.path, "type": int, "required": True},
                    },
                    "HEAD": {
                        "w": {"name": "w", "location": ParameterLocation.path, "type": int, "required": True},
                    },
                },
                id="http_function",
            ),
            pytest.param(
                "http_endpoint",
                {
                    "GET": {
                        "w": {"name": "w", "location": ParameterLocation.path, "type": int, "required": True},
                    },
                    "HEAD": {
                        "w": {"name": "w", "location": ParameterLocation.path, "type": int, "required": True},
                    },
                },
                id="http_endpoint",
            ),
            pytest.param(
                "websocket_function",
                {
                    "WEBSOCKET": {
                        "w": {"name": "w", "location": ParameterLocation.path, "type": int, "required": True},
                    }
                },
                id="websocket_function",
            ),
            pytest.param(
                "websocket_endpoint",
                {
                    "WEBSOCKET_CONNECT": {},
                    "WEBSOCKET_RECEIVE": {
                        "w": {"name": "w", "location": ParameterLocation.path, "type": int, "required": True},
                    },
                    "WEBSOCKET_DISCONNECT": {},
                },
                id="websocket_endpoint",
            ),
        ),
        indirect=["route"],
    )
    def test_path(self, route, expected_params):
        assert route.parameters.path == {
            method: {k: Parameter(**param) for k, param in params.items()} for method, params in expected_params.items()
        }

    @pytest.mark.parametrize(
        ["route", "expected_params"],
        (
            pytest.param(
                "http_function",
                {
                    "GET": {"name": "z", "location": ParameterLocation.body, "type": None},
                    "HEAD": {"name": "z", "location": ParameterLocation.body, "type": None},
                },
                id="http_function",
            ),
            pytest.param(
                "http_endpoint",
                {
                    "GET": {"name": "z", "location": ParameterLocation.body, "type": None},
                    "HEAD": {"name": "z", "location": ParameterLocation.body, "type": None},
                },
                id="http_endpoint",
            ),
            pytest.param(
                "websocket_function",
                {
                    "WEBSOCKET": {"name": "z", "location": ParameterLocation.body, "type": None},
                },
                id="websocket_function",
            ),
            pytest.param(
                "websocket_endpoint",
                {
                    "WEBSOCKET_CONNECT": None,
                    "WEBSOCKET_RECEIVE": {"name": "z", "location": ParameterLocation.body, "type": None},
                    "WEBSOCKET_DISCONNECT": None,
                },
                id="websocket_endpoint",
            ),
        ),
        indirect=["route"],
    )
    def test_body(self, route, expected_params, foo_schema):
        expected_params = {
            k: Parameter(**{**param, "type": t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)]})
            if param
            else None
            for k, param in expected_params.items()
        }
        assert route.parameters.body == expected_params

    @pytest.mark.parametrize(
        ["route", "expected_params"],
        (
            pytest.param(
                "http_function",
                {
                    "GET": {"name": "_return", "location": ParameterLocation.response, "type": True},
                    "HEAD": {"name": "_return", "location": ParameterLocation.response, "type": True},
                },
                id="http_function",
            ),
            pytest.param(
                "http_endpoint",
                {
                    "GET": {"name": "_return", "location": ParameterLocation.response, "type": True},
                    "HEAD": {"name": "_return", "location": ParameterLocation.response, "type": True},
                },
                id="http_endpoint",
            ),
            pytest.param(
                "websocket_function",
                {
                    "WEBSOCKET": {"name": "_return", "location": ParameterLocation.response, "type": None},
                },
                id="websocket_function",
            ),
            pytest.param(
                "websocket_endpoint",
                {
                    "WEBSOCKET_CONNECT": {"name": "_return", "location": ParameterLocation.response, "type": None},
                    "WEBSOCKET_RECEIVE": {"name": "_return", "location": ParameterLocation.response, "type": None},
                    "WEBSOCKET_DISCONNECT": {"name": "_return", "location": ParameterLocation.response, "type": None},
                },
                id="websocket_endpoint",
            ),
        ),
        indirect=["route"],
    )
    def test_response(self, route, expected_params, foo_schema):
        expected_params = {
            k: Parameter(
                **{
                    **param,
                    "type": t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(foo_schema)]
                    if param["type"]
                    else None,
                }
            )
            for k, param in expected_params.items()
        }
        assert route.parameters.response == expected_params



import string
import typing as t

import pydantic

import flama
from flama import Flama


class Puppy(pydantic.BaseModel):
    id: int
    name: str
    age: int

    @pydantic.field_validator("age")
    def minimum_age_validation(cls, v):
        if v < 0:
            raise ValueError("Age must be positive")

        return v


PUPPIES: list[Puppy] = [Puppy(id=1, name="Canna", age=7), Puppy(id=2, name="Sandy", age=12)]


app = Flama(
    openapi={
        "info": {
            "title": "Puppy Register",  # API title
            "version": "0.1",  # API version
            "description": "A register of puppies",  # API description
        }
    },
)


@app.route("/number/", methods=["GET"], pagination="page_number")
def numbers(**kwargs):
    """
    tags:
        - numbers.
    summary:
        A sequence of numbers.
    description:
        A sequence of numbers that uses page-number pagination.
    responses:
        200:
            description: Sequence of numbers.
    """
    return list(range(100))


@app.route("/alphabet/", methods=["GET"], pagination="limit_offset")
def alphabet(**kwargs):
    """
    tags:
        - alphabet.
    summary:
        A sequence of alphabet letters.
    description:
        A sequence of alphabet letters that uses limit-offset pagination.
    responses:
        200:
            description: Sequence of alphabet letters.
    """
    return list(string.ascii_lowercase)


@app.route("/puppy/", methods=["GET"], pagination="page_number")
def puppies(name: t.Optional[str] = None, **kwargs) -> list[Puppy]:
    """
    tags:
        - puppy
    summary:
        List puppies.
    description:
        List the puppies collection. There is an optional query parameter that
        specifies a name for filtering the collection based on it.
    responses:
        200:
            description: List puppies.
    """
    result = PUPPIES
    if name:
        result = [x for x in result if x.name == name]

    return result


if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)



import abc
import dataclasses
import functools
import io
import json
import typing as t

from flama.cli.config.app import App
from flama.cli.config.uvicorn import Uvicorn

__all__ = ["Config", "ExampleConfig", "options"]


@dataclasses.dataclass
class Config:
    app: App = dataclasses.field(
        default_factory=lambda: App.build({"title": "API title", "version": "0.1.0", "description": "API description"})
    )
    server: Uvicorn = dataclasses.field(default_factory=Uvicorn)

    @classmethod
    def from_dict(cls, data: dict[str, t.Any]) -> "Config":
        return cls(
            **{**data, "app": App.build(data["app"]), "server": Uvicorn(**data["server"])}  # type: ignore[arg-type]
        )

    def to_dict(self) -> dict[str, t.Any]:
        return dataclasses.asdict(self)

    @classmethod
    def loads(cls, data: str) -> "Config":
        return cls.from_dict(json.loads(data))

    @classmethod
    def load(cls, fs: io.StringIO) -> "Config":
        return cls.loads(fs.read())

    def dumps(self) -> str:
        return json.dumps(self.to_dict())

    def dump(self, fs: io.StringIO) -> None:
        fs.write(self.dumps())

    @classmethod
    def dump_example(cls, type: str) -> dict[str, t.Any]:
        result = cls().to_dict()
        if type == "simple":
            result["server"] = {k: v for k, v in result["server"] if k in ("host", "port")}
        return result

    def run(self) -> None:
        with self.app.context as app_context:
            self.server.app_dir = app_context.dir
            self.server.run(app_context.app)


class ExampleConfig(metaclass=abc.ABCMeta):
    config = Config()

    @classmethod
    def build(cls, mode: str) -> "ExampleConfig":
        return {"simple": SimpleExample(), "full": FullExample()}[mode]

    @classmethod
    @abc.abstractmethod
    def dumps(cls) -> str: ...


class SimpleExample(ExampleConfig):
    @classmethod
    def dumps(cls) -> str:
        return json.dumps(
            {
                **cls.config.to_dict(),
                "server": {k: v for k, v in cls.config.to_dict()["server"].items() if k in ("host", "port")},
            }
        )


class FullExample(ExampleConfig):
    @classmethod
    def dumps(cls):
        return json.dumps(cls.config.to_dict())


def options(command: t.Callable) -> t.Callable:
    """Decorate a click command with all Config options.

    :param command: Command to be decorated.
    :return: Decorated command.
    """

    @functools.wraps(command)
    def _inner(*args, **kwargs):
        command(*args, **kwargs)

    return _inner



import uuid

import pytest

from flama import Flama
from flama.authentication.components import AccessTokenComponent
from flama.authentication.middleware import AuthenticationMiddleware
from flama.middleware import Middleware

TOKENS = {
    "permission": b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJwZXJtaXNzaW9ucyI6IFsiZmxhbWEudGVzdC5hdXRoI"
    b"l19LCAiaWF0IjogMTY5ODQxMjkzMX0=.NLhM8r2g1I_oHG0zAAsRqDAuwPVvzI95Lnz2K7uupmo=",
    "role": b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJyb2xlcyI6IHsiZm9vIjogWyJmbGFtYS50ZXN0LmF1dGgiXX1"
    b"9LCAiaWF0IjogMTY5ODQxMjI5OH0=.WK2PwipkiLATHsKwIsiljS_31h0-T6U0hZzoI62Skiw=",
    "empty": b"eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjoge30sICJpYXQiOiAxNjk4NDEzMDg3fQ==.xPwYcCE0Tq6UxVbhW"
    b"RGqT8vRliJxHAqHs12X0oHE1Vg=",
}


class TestCaseAuthenticationMiddleware:
    @pytest.fixture(scope="function")
    def secret(self):
        return uuid.UUID(int=0)

    @pytest.fixture(scope="function")
    def app(self, secret):
        return Flama(
            schema=None,
            docs=None,
            components=[AccessTokenComponent(secret=secret.bytes)],
            middleware=[Middleware(AuthenticationMiddleware, ignored=[r"/ignored.*"])],
        )

    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/no-auth/")
        def no_auth():
            return {"foo": "no-auth"}

        @app.route("/auth/", tags={"permissions": ["flama.test.auth"]})
        def auth():
            return {"foo": "auth"}

        @app.route("/ignored/", tags={"permissions": ["flama.test.auth"]})
        def ignored():
            return {"foo": "ignored"}

    @pytest.fixture(scope="function")
    def headers(self, request):
        if request.param is None:
            return None

        try:
            return {"access_token": f"Bearer {TOKENS[request.param].decode()}"}
        except KeyError:
            raise ValueError(f"Invalid token {request.param}")

    @pytest.fixture(scope="function")
    def cookies(self, request):
        if request.param is None:
            return None

        try:
            return {"access_token": TOKENS[request.param].decode()}
        except KeyError:
            raise ValueError(f"Invalid token {request.param}")

    @pytest.mark.parametrize(
        ["path", "method", "headers", "cookies", "status_code", "result"],
        (
            pytest.param("/auth/", "get", "permission", None, 200, {"foo": "auth"}, id="auth_header_token_permission"),
            pytest.param("/auth/", "get", "role", None, 200, {"foo": "auth"}, id="auth_header_token_role"),
            pytest.param(
                "/auth/",
                "get",
                "empty",
                None,
                403,
                {"detail": "Insufficient permissions", "error": None, "status_code": 403},
                id="auth_header_token_empty",
            ),
            pytest.param("/auth/", "get", None, "permission", 200, {"foo": "auth"}, id="auth_cookie_token_permission"),
            pytest.param("/auth/", "get", None, "role", 200, {"foo": "auth"}, id="auth_cookie_token_role"),
            pytest.param(
                "/auth/",
                "get",
                None,
                "empty",
                403,
                {"detail": "Insufficient permissions", "error": None, "status_code": 403},
                id="auth_cookie_token_empty",
            ),
            pytest.param(
                "/auth/",
                "get",
                None,
                None,
                401,
                {"detail": "Unauthorized", "error": None, "status_code": 401},
                id="auth_no_token",
            ),
            pytest.param(
                "/no-auth/", "get", "permission", None, 200, {"foo": "no-auth"}, id="no_auth_header_token_permission"
            ),
            pytest.param("/no-auth/", "get", "role", None, 200, {"foo": "no-auth"}, id="no_auth_header_token_role"),
            pytest.param(
                "/no-auth/", "get", None, "permission", 200, {"foo": "no-auth"}, id="no_auth_cookie_token_permission"
            ),
            pytest.param("/no-auth/", "get", None, "role", 200, {"foo": "no-auth"}, id="no_auth_cookie_token_role"),
            pytest.param("/no-auth/", "get", None, None, 200, {"foo": "no-auth"}, id="no_auth_no_token"),
            pytest.param("/ignored/", "get", None, None, 200, {"foo": "ignored"}, id="no_auth_no_token"),
            pytest.param(
                "/not-found/",
                "get",
                None,
                None,
                404,
                {"detail": "Not Found", "error": "HTTPException", "status_code": 404},
                id="not_found",
            ),
            pytest.param(
                "/auth/",
                "post",
                None,
                None,
                405,
                {"detail": "Method Not Allowed", "error": "HTTPException", "status_code": 405},
                id="method_not_allowed",
            ),
        ),
        indirect=["headers", "cookies"],
    )
    async def test_request(self, client, path, method, headers, cookies, status_code, result):
        client.headers = headers
        client.cookies = cookies
        response = await client.request(method, path)

        assert response.status_code == status_code
        assert response.json() == result



import builtins
import dataclasses
import enum
import typing as t

from flama import compat, schemas, types
from flama.injection.resolver import Parameter as InjectionParameter

__all__ = ["Field", "Schema", "Parameter", "Parameters"]


UNKNOWN = t.TypeVar("UNKNOWN")


class ParameterLocation(compat.StrEnum):  # PORT: Replace compat when stop supporting 3.10
    query = enum.auto()
    path = enum.auto()
    body = enum.auto()
    response = enum.auto()


@dataclasses.dataclass(frozen=True)
class Field:
    name: str
    type: type
    nullable: bool = dataclasses.field(init=False)
    field: t.Any = dataclasses.field(hash=False, init=False, compare=False)
    multiple: t.Optional[bool] = dataclasses.field(hash=False, compare=False, default=None)
    required: bool = True
    default: t.Any = InjectionParameter.empty

    def __post_init__(self) -> None:
        object.__setattr__(self, "nullable", type(None) in t.get_args(self.type) or self.default is None)

        field_type = t.get_args(self.type)[0] if t.get_origin(self.type) in (t.Union, compat.UnionType) else self.type

        if not Schema.is_schema(field_type) and self.multiple is None:
            object.__setattr__(self, "multiple", t.get_origin(self.type) is list)

        object.__setattr__(
            self,
            "field",
            schemas.adapter.build_field(
                self.name,
                field_type,
                nullable=self.nullable,
                required=self.required,
                default=self.default,
                multiple=bool(self.multiple),
            ),
        )

    @classmethod
    def from_parameter(cls, parameter: InjectionParameter) -> "Field":
        return cls(
            parameter.name,
            parameter.annotation,
            required=parameter.default is InjectionParameter.empty,
            default=parameter.default
            if parameter.default is not InjectionParameter.empty
            else InjectionParameter.empty,
        )

    @classmethod
    def is_field(cls, obj: t.Any) -> bool:
        return schemas.adapter.is_field(obj)

    @classmethod
    def is_http_valid_type(cls, type_: builtins.type) -> bool:
        origin = t.get_origin(type_)
        args = t.get_args(type_)
        NoneType = type(None)

        return (
            (type_ in types.PARAMETERS_TYPES)
            or (
                origin in (t.Union, compat.UnionType)
                and len(args) == 2
                and args[0] in types.PARAMETERS_TYPES
                and args[1] is NoneType
            )
            or (origin is list and args[0] in types.PARAMETERS_TYPES)
        )

    @property
    def json_schema(self) -> types.JSONSchema:
        return schemas.adapter.to_json_schema(self.field)


@dataclasses.dataclass(frozen=True)
class Schema:
    schema: t.Any = dataclasses.field(hash=False, compare=False)

    @classmethod
    def from_type(cls, type_: t.Optional[type]) -> "Schema":
        if schemas.is_schema(type_):
            schema = schemas.get_schema_metadata(type_).schema

            if schemas.is_schema_partial(type_):
                schema = schemas.adapter.build_schema(
                    name=schemas.adapter.name(schema, prefix="Partial").rsplit(".", 1)[1], schema=schema, partial=True
                )
        elif t.get_origin(type_) in (list, tuple, set):
            return cls.from_type(t.get_args(type_)[0])
        else:
            schema = type_

        if not schemas.adapter.is_schema(schema):
            raise ValueError("Wrong schema type")

        return cls(schema=schema)

    @classmethod
    def build(
        cls,
        name: t.Optional[str] = None,
        schema: t.Any = None,
        fields: t.Optional[list[Field]] = None,
    ) -> "Schema":
        return cls(
            schema=schemas.adapter.build_schema(
                name=name, schema=schema, fields={f.name: f.field for f in (fields or [])}
            ),
        )

    @classmethod
    def is_schema(cls, obj: t.Any) -> bool:
        return schemas.adapter.is_schema(obj)

    @property
    def name(self) -> str:
        return schemas.adapter.name(self.schema)

    def _fix_ref(self, value: str, refs: dict[str, str]) -> str:
        try:
            prefix, name = value.rsplit("/", 1)
            return f"{prefix}/{refs[name]}"
        except KeyError:
            return value

    def _replace_json_schema_refs(self, schema: types.JSONField, refs: dict[str, str]) -> types.JSONField:
        if isinstance(schema, dict):
            return {
                k: self._fix_ref(t.cast(str, v), refs) if k == "$ref" else self._replace_json_schema_refs(v, refs)
                for k, v in schema.items()
            }

        if isinstance(schema, (list, tuple, set)):
            return [self._replace_json_schema_refs(x, refs) for x in schema]

        return schema

    def json_schema(self, names: dict[int, str]) -> types.JSONSchema:
        return t.cast(
            types.JSONSchema,
            self._replace_json_schema_refs(
                schemas.adapter.to_json_schema(self.schema),
                {Schema(x).name.rsplit(".", 1)[1]: names[id(Schema(x).unique_schema)] for x in self.nested_schemas()},
            ),
        )

    @property
    def unique_schema(self) -> t.Any:
        return schemas.adapter.unique_schema(self.schema)

    @property
    def fields(self) -> dict[str, tuple[t.Any, t.Any]]:
        return schemas.adapter.schema_fields(self.unique_schema)

    def nested_schemas(self, schema: t.Any = UNKNOWN) -> list[t.Any]:
        if schema == UNKNOWN:
            return self.nested_schemas(self)

        if schemas.adapter.is_schema(schema):
            return [schemas.adapter.unique_schema(schema)]

        if t.get_origin(schema) in (t.Union, compat.UnionType):
            return [x for field in t.get_args(schema) for x in self.nested_schemas(field)]

        if isinstance(schema, (list, tuple, set)):
            return [x for field in schema for x in self.nested_schemas(field)]

        if isinstance(schema, dict):
            return [x for field in schema.values() for x in self.nested_schemas(field)]

        if isinstance(schema, Schema):
            return [x for field_type, _ in schema.fields.values() for x in self.nested_schemas(field_type)]

        return []

    @t.overload
    def validate(self, values: None, *, partial: bool = False) -> dict[str, t.Any]: ...

    @t.overload
    def validate(self, values: dict[str, t.Any], *, partial: bool = False) -> dict[str, t.Any]: ...

    @t.overload
    def validate(self, values: list[dict[str, t.Any]], *, partial: bool = False) -> list[dict[str, t.Any]]: ...

    def validate(self, values: t.Union[dict[str, t.Any], list[dict[str, t.Any]], None], *, partial=False):
        if isinstance(values, (list, tuple)):
            return [schemas.adapter.validate(self.schema, value, partial=partial) for value in values]

        return schemas.adapter.validate(self.schema, values or {}, partial=partial)

    @t.overload
    def load(self, values: dict[str, t.Any]) -> t.Any: ...

    @t.overload
    def load(self, values: list[dict[str, t.Any]]) -> list[t.Any]: ...

    def load(self, values):
        if isinstance(values, (list, tuple)):
            return [schemas.adapter.load(self.schema, value) for value in values]

        return schemas.adapter.load(self.schema, values)

    @t.overload
    def dump(self, values: dict[str, t.Any]) -> dict[str, t.Any]: ...

    @t.overload
    def dump(self, values: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]: ...

    def dump(self, values):
        if isinstance(values, (list, tuple)):
            return [schemas.adapter.dump(self.schema, value) for value in values]

        return schemas.adapter.dump(self.schema, values)


@dataclasses.dataclass(frozen=True)
class Parameter:
    name: str
    location: ParameterLocation
    type: t.Any
    required: bool = True
    default: t.Any = InjectionParameter.empty
    nullable: bool = dataclasses.field(init=False)
    multiple: bool = dataclasses.field(init=False)
    schema: "Schema" = dataclasses.field(hash=False, init=False, compare=False)
    field: "Field" = dataclasses.field(hash=False, init=False, compare=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "nullable", type(None) in t.get_args(self.type) or self.default is None)

        try:
            schema = Schema.from_type(self.type)
            field = None
        except ValueError:
            if self.type in (None, InjectionParameter.empty):
                schema = Schema(schema=None)
                field = None
            else:
                schema = None
                field = Field(self.name, self.type, required=self.required, default=self.default)

        object.__setattr__(self, "schema", schema)
        object.__setattr__(self, "field", field)
        object.__setattr__(self, "multiple", t.get_origin(self.type) in (list, tuple, set, frozenset))

    @classmethod
    def build(cls, type_: str, parameter: InjectionParameter):
        return {
            "path": cls._build_path_parameter,
            "query": cls._build_query_parameter,
            "body": cls._build_body_parameter,
            "response": cls._build_response_parameter,
        }[type_](parameter)

    @classmethod
    def _build_path_parameter(cls, parameter: InjectionParameter) -> "Parameter":
        return cls(
            name=parameter.name,
            location=ParameterLocation.path,
            type=parameter.annotation if parameter.annotation is not parameter.empty else str,
        )

    @classmethod
    def _build_query_parameter(cls, parameter: InjectionParameter) -> "Parameter":
        return cls(
            name=parameter.name,
            location=ParameterLocation.query,
            type=parameter.annotation if parameter.annotation is not parameter.empty else str,
            required=parameter.default is InjectionParameter.empty,
            default=parameter.default,
        )

    @classmethod
    def _build_body_parameter(cls, parameter: InjectionParameter) -> "Parameter":
        return cls(name=parameter.name, location=ParameterLocation.body, type=parameter.annotation)

    @classmethod
    def _build_response_parameter(cls, parameter: InjectionParameter) -> "Parameter":
        return cls(name=parameter.name, location=ParameterLocation.response, type=parameter.annotation)


Parameters = dict[str, Parameter]



import http.cookiejar

import pytest

from flama import types


class TestCaseMethodComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/method/", methods=["GET", "POST"])
        def get_method(method: types.Method):
            return {"method": str(method)}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param("/method/", "get", {"method": "GET"}, id="get"),
            pytest.param("/method/", "post", {"method": "POST"}, id="post"),
        ],
    )
    async def test_method(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCaseURLComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/url/")
        def get_url(url: types.URL):
            return {"url": url.url, "components": url.components}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param(
                "http://example.com/url/",
                "get",
                {
                    "url": "http://example.com/url/",
                    "components": {
                        "scheme": "http",
                        "netloc": "example.com",
                        "path": "/url/",
                        "query": "",
                        "fragment": "",
                        "params": "",
                    },
                },
                id="http",
            ),
            pytest.param(
                "https://example.com/url/",
                "get",
                {
                    "url": "https://example.com/url/",
                    "components": {
                        "scheme": "https",
                        "netloc": "example.com",
                        "path": "/url/",
                        "query": "",
                        "fragment": "",
                        "params": "",
                    },
                },
                id="https",
            ),
            pytest.param(
                "http://example.com:123/url/",
                "get",
                {
                    "url": "http://example.com:123/url/",
                    "components": {
                        "scheme": "http",
                        "netloc": "example.com:123",
                        "path": "/url/",
                        "query": "",
                        "fragment": "",
                        "params": "",
                    },
                },
                id="http_port",
            ),
            pytest.param(
                "https://example.com:123/url/",
                "get",
                {
                    "url": "https://example.com:123/url/",
                    "components": {
                        "scheme": "https",
                        "netloc": "example.com:123",
                        "path": "/url/",
                        "query": "",
                        "fragment": "",
                        "params": "",
                    },
                },
                id="https_port",
            ),
            pytest.param(
                "http://example.com/url/?a=1",
                "get",
                {
                    "url": "http://example.com/url/?a=1",
                    "components": {
                        "scheme": "http",
                        "netloc": "example.com",
                        "path": "/url/",
                        "query": "a=1",
                        "fragment": "",
                        "params": "",
                    },
                },
                id="query_param",
            ),
        ],
    )
    async def test_url(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCaseSchemeComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/scheme/")
        def get_scheme(scheme: types.Scheme):
            return {"scheme": scheme}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param("http://example.com/scheme/", "get", {"scheme": "http"}, id="http"),
            pytest.param("https://example.com/scheme/", "get", {"scheme": "https"}, id="https"),
        ],
    )
    async def test_scheme(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCaseHostComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/host/")
        def get_host(host: types.Host):
            return {"host": host}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param("http://example.com/host/", "get", {"host": "example.com"}, id="host"),
            pytest.param("http://example.com:123/host/", "get", {"host": "example.com"}, id="host_and_port"),
        ],
    )
    async def test_host(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCasePortComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/port/")
        def get_port(port: types.Port):
            return {"port": port}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param("http://example.com/port/", "get", {"port": None}, id="http_default_implicit"),
            pytest.param("https://example.com/port/", "get", {"port": None}, id="https_default_implicit"),
            pytest.param("http://example.com:80/port/", "get", {"port": None}, id="http_default_explicit"),
            pytest.param("https://example.com:443/port/", "get", {"port": None}, id="https_default_explicit"),
            pytest.param("http://example.com:123/port/", "get", {"port": 123}, id="http_custom"),
            pytest.param("https://example.com:123/port/", "get", {"port": 123}, id="https_custom"),
        ],
    )
    async def test_port(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCasePathComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/path/")
        def get_path(path: types.Path):
            return {"path": path}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param("/path/", "get", {"path": "/path/"}, id="path"),
        ],
    )
    async def test_path(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCasePathParamsComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/path_params/{foo:int}/subpath/{bar:str}/")
        def get_path_params(path_params: types.PathParams):
            return {"path_params": path_params}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param(
                "/path_params/1/subpath/bar/", "get", {"path_params": {"foo": 1, "bar": "bar"}}, id="path_params"
            ),
        ],
    )
    async def test_path_params(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCaseQueryStringComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/query_string/")
        def get_query_string(query_string: types.QueryString):
            return {"query_string": query_string}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param("/query_string/", "get", {"query_string": ""}, id="empty"),
            pytest.param("/query_string/?a=1&a=2&b=3", "get", {"query_string": "a=1&a=2&b=3"}, id="params"),
        ],
    )
    async def test_query_string(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCaseQueryParamsComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/query_params/")
        def get_query_params(query_string: types.QueryString, query_params: types.QueryParams):
            return {"query_params": dict(query_params)}

    @pytest.mark.parametrize(
        ["path", "method", "expected"],
        [
            pytest.param("/query_params/", "get", {"query_params": {}}, id="empty"),
            pytest.param("/query_params/?a=1&a=2&b=3", "get", {"query_params": {"a": "2", "b": "3"}}, id="params"),
        ],
    )
    async def test_query_params(self, client, path, method, expected):
        response = await client.request(method, path)
        assert response.json() == expected


class TestCaseHeadersComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/headers/", methods=["GET", "POST"])
        def get_headers(headers: types.Headers):
            return {"headers": dict(headers)}

    @pytest.mark.parametrize(
        "path,method,request_kwargs,expected",
        [
            pytest.param(
                "http://example.com/headers/",
                "get",
                {},
                {
                    "headers": {
                        "accept": "*/*",
                        "accept-encoding": "gzip, deflate",
                        "connection": "keep-alive",
                        "host": "example.com",
                    }
                },
                id="default",
            ),
            pytest.param(
                "http://example.com/headers/",
                "get",
                {"headers": {"X-Example-Header": "example"}},
                {
                    "headers": {
                        "accept": "*/*",
                        "accept-encoding": "gzip, deflate",
                        "connection": "keep-alive",
                        "host": "example.com",
                        "x-example-header": "example",
                    }
                },
                id="custom",
            ),
            pytest.param(
                "http://example.com/headers/",
                "get",
                {"data": {"a": 1}},
                {
                    "headers": {
                        "accept": "*/*",
                        "accept-encoding": "gzip, deflate",
                        "connection": "keep-alive",
                        "content-length": "3",
                        "content-type": "application/x-www-form-urlencoded",
                        "host": "example.com",
                    }
                },
                id="content_length",
            ),
        ],
    )
    async def test_headers(self, client, path, method, request_kwargs, expected):
        response = await client.request(method, path, **dict(request_kwargs))
        response_json = response.json()
        del response_json["headers"]["user-agent"]

        assert response_json == expected


class TestCaseCookiesComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/cookies/", methods=["GET", "POST"])
        def get_cookies(cookies: types.Cookies):
            return {"cookies": dict(cookies)}

    @pytest.mark.parametrize(
        ["path", "method", "cookies", "expected"],
        [
            pytest.param(
                "http://example.com/cookies/",
                "get",
                {},
                {"cookies": {}},
                id="default",
            ),
            pytest.param(
                "http://example.com/cookies/",
                "get",
                [
                    http.cookiejar.Cookie(
                        version=0,
                        name="foo",
                        value="bar",
                        port=None,
                        port_specified=False,
                        domain="",
                        domain_specified=False,
                        domain_initial_dot=False,
                        path="/",
                        path_specified=True,
                        secure=False,
                        expires=None,
                        discard=True,
                        comment=None,
                        comment_url=None,
                        rest={"HttpOnly": ""},
                        rfc2109=False,
                    )
                ],
                {
                    "cookies": {
                        "foo": {
                            "value": "bar",
                            "expires": "",
                            "path": "",
                            "comment": "",
                            "domain": "",
                            "max-age": "",
                            "secure": "",
                            "httponly": "",
                            "version": "",
                            "samesite": "",
                        }
                    }
                },
                id="cookie",
            ),
            pytest.param(
                "http://example.com/cookies/",
                "get",
                [
                    http.cookiejar.Cookie(
                        version=0,
                        name="foo",
                        value="bar",
                        port=None,
                        port_specified=False,
                        domain="",
                        domain_specified=False,
                        domain_initial_dot=False,
                        path="/",
                        path_specified=True,
                        secure=True,
                        expires=None,
                        discard=True,
                        comment=None,
                        comment_url=None,
                        rest={"HttpOnly": "true"},
                        rfc2109=False,
                    )
                ],
                {"cookies": {}},  # Cannot get cookie because secure and no https
                id="cookie_secure",
            ),
        ],
    )
    async def test_cookies(self, client, path, method, cookies, expected):
        cookies_jar = http.cookiejar.CookieJar()
        for cookie in cookies:
            cookies_jar.set_cookie(cookie)
        client.cookies = cookies_jar
        response = await client.request(method, path)
        response_json = response.json()

        assert response_json == expected


class TestCaseBodyComponent:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/body/", methods=["GET", "POST"])
        def get_body(body: types.Body):
            return {"body": body.decode("utf-8")}

    @pytest.mark.parametrize(
        ["path", "method", "request_kwargs", "expected"],
        [
            pytest.param("/body/", "post", {"content": "content"}, {"body": "content"}, id="default"),
        ],
    )
    async def test_body(self, client, path, method, request_kwargs, expected):
        response = await client.request(method, path, **dict(request_kwargs))
        assert response.json() == expected



import uuid
from unittest.mock import Mock, call, patch

import pytest
import sqlalchemy
from sqlalchemy.ext.asyncio import AsyncConnection

from flama import Flama
from flama.ddd import exceptions
from flama.ddd.repositories.sqlalchemy import SQLAlchemyRepository, SQLAlchemyTableManager, SQLAlchemyTableRepository
from flama.sqlalchemy import SQLAlchemyModule
from tests.utils import SQLAlchemyContext


@pytest.fixture(scope="function")
def app():
    return Flama(schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")})


@pytest.fixture(scope="function")
async def connection(client):
    # Exactly the same behavior as 'async with worker'
    connection_: AsyncConnection = client.app.sqlalchemy.engine.connect()
    await connection_.__aenter__()
    transaction = connection_.begin()
    await transaction.__aenter__()

    yield connection_

    await transaction.__aexit__(None, None, None)
    await connection_.__aexit__(None, None, None)


@pytest.fixture(scope="function")
def tables(app):
    return {
        "single": sqlalchemy.Table(
            "repository_table_single_pk",
            app.sqlalchemy.metadata,
            sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True, autoincrement=True),
            sqlalchemy.Column("name", sqlalchemy.String, nullable=False),
        ),
        "composed": sqlalchemy.Table(
            "repository_table_composed_pk",
            app.sqlalchemy.metadata,
            sqlalchemy.Column("id_first", sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column("id_second", sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column("name", sqlalchemy.String, nullable=False),
        ),
    }


@pytest.fixture(scope="function")
async def table(client, tables):
    table = tables["single"]

    async with SQLAlchemyContext(client.app, [table]):
        yield table


class TestCaseSQLAlchemyRepository:
    @pytest.fixture(scope="function")
    def connection(self):
        return Mock(spec=AsyncConnection)

    async def test_init(self, connection):
        class Repository(SQLAlchemyRepository): ...

        repository = Repository(connection)

        assert repository._connection == connection

    def test_eq(self, connection):
        assert SQLAlchemyRepository(connection) == SQLAlchemyRepository(connection)


class TestCaseSQLAlchemyTableManager:
    @pytest.fixture(scope="function")
    def table_manager(self, table, connection):
        return SQLAlchemyTableManager(table, connection)

    async def test_init(self, table, connection):
        table_manager = SQLAlchemyTableManager(table, connection)

        assert table_manager._connection == connection
        assert table_manager.table == table

    def test_eq(self, table, connection):
        assert SQLAlchemyTableManager(table, connection) == SQLAlchemyTableManager(table, connection)

    @pytest.mark.parametrize(
        ["data", "result", "exception"],
        (
            pytest.param(
                [{"name": "foo"}],
                [{"id": 1, "name": "foo"}],
                None,
                id="single",
            ),
            pytest.param(
                [{"name": "foo"}, {"name": "bar"}],
                [{"id": 1, "name": "foo"}, {"id": 2, "name": "bar"}],
                None,
                id="multiple",
            ),
            pytest.param(
                [{"name": None}],
                None,
                exceptions.IntegrityError,
                id="integrity_error",
            ),
        ),
        indirect=["exception"],
    )
    async def test_create(self, table_manager, data, result, exception):
        with exception:
            assert await table_manager.create(*data) == result

    @pytest.mark.parametrize(
        ["clauses", "filters", "result", "exception"],
        (
            pytest.param(
                [],
                {"id": 1},
                {"id": 1, "name": "foo"},
                None,
                id="ok",
            ),
            pytest.param(
                [],
                {"id": 0},
                None,
                exceptions.NotFoundError(resource="repository_table_single_pk"),
                id="not_found",
            ),
            pytest.param(
                [lambda x: x.ilike("fo%")],
                {},
                None,
                exceptions.MultipleRecordsError(resource="repository_table_single_pk"),
                id="multiple_results",
            ),
        ),
        indirect=["exception"],
    )
    async def test_retrieve(self, clauses, filters, result, exception, table, table_manager):
        await table_manager.create({"name": "foo"}, {"name": "foo"})

        with exception:
            assert await table_manager.retrieve(*[c(table.c["name"]) for c in clauses], **filters) == result

    @pytest.mark.parametrize(
        ["clauses", "filters", "data", "result", "exception"],
        (
            pytest.param(
                [],
                {"id": 1},
                {"name": "bar"},
                [{"id": 1, "name": "bar"}],
                None,
                id="ok",
            ),
            pytest.param(
                [],
                {"id": 0},
                {"name": "bar"},
                [],
                None,
                id="not_found",
            ),
            pytest.param(
                [],
                {"id": 1},
                {"name": None},
                [{"id": 1, "name": None}],
                exceptions.IntegrityError,
                id="integrity_error",
            ),
            pytest.param(
                [lambda x: x.ilike("fo%")],
                {},
                {"name": "bar"},
                [{"id": 1, "name": "bar"}, {"id": 2, "name": "bar"}],
                None,
                id="multiple_results",
            ),
        ),
        indirect=["exception"],
    )
    async def test_update(self, clauses, filters, data, result, exception, table, table_manager):
        await table_manager.create({"name": "foo"}, {"name": "foo"})

        with exception:
            assert await table_manager.update(data, *[c(table.c["name"]) for c in clauses], **filters) == result

    @pytest.mark.parametrize(
        ["clauses", "filters", "exception"],
        (
            pytest.param(
                [],
                {"id": 1},
                None,
                id="ok",
            ),
            pytest.param(
                [],
                {"id": 0},
                exceptions.NotFoundError(resource="repository_table_single_pk"),
                id="not_found",
            ),
            pytest.param(
                [lambda x: x.ilike("fo%")],
                {},
                exceptions.MultipleRecordsError,
                id="multiple_results",
            ),
        ),
        indirect=["exception"],
    )
    async def test_delete(self, clauses, filters, exception, table, table_manager):
        await table_manager.create({"name": "foo"}, {"name": "foo"})

        with exception:
            await table_manager.delete(*[c(table.c["name"]) for c in clauses], **filters)

    @pytest.mark.parametrize(
        ["clauses", "order_by", "order_direction", "filters", "result"],
        (
            pytest.param([], None, None, {}, [{"id": 1, "name": "foo"}, {"id": 2, "name": "bar"}], id="all"),
            pytest.param([lambda x: x.ilike("fo%")], None, None, {}, [{"id": 1, "name": "foo"}], id="clauses"),
            pytest.param([], "name", "asc", {}, [{"id": 2, "name": "bar"}, {"id": 1, "name": "foo"}], id="order"),
            pytest.param([], None, None, {"name": "foo"}, [{"id": 1, "name": "foo"}], id="filters"),
        ),
    )
    async def test_list(self, clauses, order_by, order_direction, filters, result, table, table_manager):
        await table_manager.create({"name": "foo"}, {"name": "bar"})

        r = [
            x
            async for x in table_manager.list(
                *[c(table.c["name"]) for c in clauses], order_by=order_by, order_direction=order_direction, **filters
            )
        ]

        assert r == result

    @pytest.mark.parametrize(
        ["clauses", "filters", "result"],
        (
            pytest.param([], {}, 2, id="all"),
            pytest.param([lambda x: x.ilike("fo%")], {}, 1, id="clauses"),
            pytest.param([], {"name": "foo"}, 1, id="filters"),
        ),
    )
    async def test_drop(self, clauses, filters, result, table, table_manager):
        await table_manager.create({"name": "foo"}, {"name": "bar"})

        r = await table_manager.drop(*[c(table.c["name"]) for c in clauses], **filters)

        assert r == result


class TestCaseSQLAlchemyTableRepository:
    @pytest.fixture(scope="function")
    def table(self):
        mock = Mock(spec=sqlalchemy.Table)
        mock.name = "foo"
        return mock

    @pytest.fixture(scope="function")
    def connection(self):
        return Mock(spec=AsyncConnection)

    @pytest.fixture(scope="function")
    def table_manager(self):
        return Mock(spec=SQLAlchemyTableManager)

    @pytest.fixture(scope="function")
    def repository(self, table, connection, table_manager):
        class Repository(SQLAlchemyTableRepository):
            _table = table

        r = Repository(connection)
        with patch.object(r, "_table_manager", table_manager):
            yield r

    async def test_init(self, table, connection):
        class Repository(SQLAlchemyTableRepository):
            _table = table

        repository = Repository(connection)

        assert repository._connection == connection
        assert repository._table_manager == SQLAlchemyTableManager(table, connection)

    def test_eq(self, table, connection):
        class Repository(SQLAlchemyTableRepository):
            _table = table

        repository = Repository(connection)
        assert repository == Repository(connection)
        assert repository != Repository(Mock(spec=AsyncConnection))

    async def test_create(self, repository, table_manager):
        data = {"foo": "bar"}

        await repository.create(data)

        assert table_manager.create.call_args_list == [call(data)]

    async def test_retrieve(self, repository, table_manager):
        id = uuid.uuid4()

        await repository.retrieve(id)

        assert table_manager.retrieve.call_args_list == [call(id)]

    async def test_update(self, repository, table_manager):
        id = uuid.uuid4()
        data = {"foo": "bar"}

        await repository.update(id, data)

        assert table_manager.update.call_args_list == [call(id, data)]

    async def test_delete(self, repository, table_manager):
        id = uuid.uuid4()

        await repository.delete(id)

        assert table_manager.delete.call_args_list == [call(id)]

    async def test_list(self, repository, table_manager):
        clauses = [Mock(), Mock()]
        order_by = "foo"
        order_direction = "desc"
        filters = {"foo": "bar"}

        repository.list(*clauses, order_by=order_by, order_direction=order_direction, **filters)

        assert table_manager.list.call_args_list == [
            call(*clauses, order_by=order_by, order_direction=order_direction, **filters)
        ]

    async def test_drop(self, repository, table_manager):
        await repository.drop()

        assert table_manager.drop.call_args_list == [call()]



import flama
from flama import Component, Flama


class Address:
    def __init__(self, address: str, zip_code: str):
        self.address = address
        self.zip_code = zip_code

    def to_dict(self):
        return {"address": self.address, "zip_code": self.zip_code}


class AddressComponent(Component):
    def resolve(self, address: str, zip_code: str) -> Address:
        return Address(address, zip_code)


class Person:
    def __init__(self, name: str, age: int, address: Address):
        self.name = name
        self.age = age
        self.address = address

    def to_dict(self):
        return {"name": self.name, "age": self.age, "address": self.address.to_dict()}


class PersonComponent(Component):
    def resolve(self, name: str, age: int, address: Address) -> Person:
        return Person(name, age, address)


app = Flama(components=[PersonComponent(), AddressComponent()])


@app.get("/foo")
def person(person: Person):
    return {"person": person.to_dict()}


if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)



import datetime
import typing as t
import uuid

import starlette.datastructures

from flama.url import URL, Path

__all__ = [
    "Method",
    "Method",
    "Scheme",
    "Host",
    "Port",
    "Path",
    "QueryString",
    "QueryParam",
    "Header",
    "Body",
    "PathParams",
    "PathParam",
    "RequestData",
    "URL",
    "Headers",
    "MutableHeaders",
    "Cookies",
    "QueryParams",
    "PARAMETERS_TYPES",
]

Method = t.NewType("Method", str)
Scheme = t.NewType("Scheme", str)
Host = t.NewType("Host", str)
Port = t.NewType("Port", int)
QueryString = t.NewType("QueryString", str)
QueryParam = t.NewType("QueryParam", str)
Header = t.NewType("Header", str)
Body = t.NewType("Body", bytes)
PathParams = t.NewType("PathParams", dict[str, t.Any])
PathParam = t.NewType("PathParam", str)
RequestData = t.NewType("RequestData", dict[str, t.Any])
Headers = starlette.datastructures.Headers
MutableHeaders = starlette.datastructures.MutableHeaders
Cookies = t.NewType("Cookies", dict[str, dict[str, str]])
QueryParams = starlette.datastructures.QueryParams

PARAMETERS_TYPES: dict[type, type] = {
    int: int,
    float: float,
    str: str,
    bool: bool,
    uuid.UUID: uuid.UUID,
    datetime.date: datetime.date,
    datetime.datetime: datetime.datetime,
    datetime.time: datetime.time,
    QueryParam: str,
    PathParam: str,
}



import pytest

from flama import types
from flama.ddd.components import WorkerComponent
from flama.ddd.repositories.base import BaseRepository
from flama.ddd.workers.base import BaseWorker
from flama.injection.resolver import Parameter


class TestCaseWorkerComponent:
    @pytest.fixture(scope="function")
    def repository(self):
        class FooRepository(BaseRepository): ...

        return FooRepository

    @pytest.fixture(scope="function")
    def worker(self, repository):
        class FooWorker(BaseWorker):
            foo: repository

            async def set_up(self): ...

            async def tear_down(self, *, rollback: bool = False): ...

            async def repository_params(self):
                return [], {}

            async def commit(self): ...

            async def rollback(self): ...

        return FooWorker()

    @pytest.fixture(scope="function")
    def component(self, worker):
        return WorkerComponent(worker)

    def test_init(self, component, worker):
        assert component.worker == worker

    @pytest.fixture(scope="function")
    def parameter_types(self, worker):
        return {
            "abstract_worker": BaseWorker,
            "implemented_worker": worker.__class__,
            "int": int,
        }

    @pytest.mark.parametrize(
        ["param_name", "param_type", "expected"],
        [
            pytest.param("foo", "implemented_worker", True, id="handle"),
            pytest.param("foo", "abstract_worker", False, id="not_handle_abstract"),
            pytest.param("foo", "int", False, id="not_handle_int"),
        ],
    )
    def test_can_handle_parameter(self, component, param_name, param_type, parameter_types, expected):
        assert component.can_handle_parameter(Parameter(param_name, parameter_types[param_type])) == expected

    def test_resolve(self, component, worker):
        class App: ...

        foo = App()
        bar = App()

        scopes = types.Scope({"app": foo, "root_app": bar})

        assert hasattr(component.worker, "_app")
        assert not component.worker._app

        resolved = component.resolve(scopes)

        assert resolved == worker
        assert hasattr(component.worker, "_app")
        assert component.worker._app == scopes["root_app"]



from unittest.mock import Mock, call, patch

import pytest

from flama import Flama, exceptions, pagination, schemas
from flama.schemas.modules import SchemaModule


class TestCaseSchemaModule:
    @pytest.fixture
    def module(self, openapi_spec):
        m = SchemaModule(openapi=openapi_spec, schema="/schema/", docs="/docs/")
        m.app = Flama()
        return m

    def test_init(self, openapi_spec):
        module = SchemaModule(openapi_spec, schema="/schema/", docs="/docs/")

        assert module.openapi == openapi_spec

    def test_register_schema(self, module, foo_schema):
        assert module.schemas == {}

        module.register_schema("foo", foo_schema)

        assert module.schemas == {"foo": foo_schema}

    def test_schema_generator(self, module):
        with patch("flama.schemas.modules.SchemaGenerator") as generator_mock:
            module.schema_generator

        assert generator_mock.call_args_list == [
            call(
                spec={"info": {"title": "Foo", "version": "1.0.0", "description": "Bar"}},
                schemas={**schemas.schemas.SCHEMAS, **pagination.paginator.schemas},
            )
        ]

    def test_schema(self, module):
        assert module.schema == {
            "openapi": "3.1.0",
            "info": {"title": "Foo", "version": "1.0.0", "description": "Bar"},
            "paths": {},
            "components": {
                "schemas": {},
                "responses": {},
                "parameters": {},
                "examples": {},
                "requestBodies": {},
                "headers": {},
                "securitySchemes": {},
                "links": {},
                "callbacks": {},
            },
        }

    @pytest.mark.parametrize(
        ["schema", "docs", "exception"],
        (
            pytest.param(
                False,
                False,
                None,
                id="no_schema_no_docs",
            ),
            pytest.param(
                True,
                False,
                None,
                id="schema_but_no_docs",
            ),
            pytest.param(
                False,
                True,
                exceptions.ApplicationError("Docs endpoint needs schema endpoint to be active"),
                id="no_schema_but_docs",
            ),
            pytest.param(
                True,
                True,
                None,
                id="schema_and_docs",
            ),
        ),
        indirect=["exception"],
    )
    def test_add_routes(self, openapi_spec, schema, docs, exception):
        with exception:
            module = SchemaModule(openapi_spec, schema=schema, docs=docs)
            module.app = Mock(Flama)
            expected_calls = []
            if schema:
                expected_calls.append(call(schema, module.schema_view, methods=["GET"], include_in_schema=False))
            if docs:
                expected_calls.append(call(docs, module.docs_view, methods=["GET"], include_in_schema=False))

            module.add_routes()

            assert module.app.add_route.call_args_list == expected_calls

    async def test_view_schema(self, client):
        response = await client.request("get", "/schema/")
        assert response.status_code == 200
        assert response.headers.get("content-type") == "application/vnd.oai.openapi"

    async def test_view_docs(self, client):
        response = await client.request("get", "/docs/")
        assert response.status_code == 200



import typing
from datetime import datetime

import flama
from flama import Flama
from flama.models import BaseModelResource, ModelComponent
from flama.models.base import Model
from flama.resources import resource_method


class MyCustomModel(Model):
    def inspect(self) -> typing.Any:
        return self.model.get_params()

    def predict(self, x: typing.Any) -> typing.Any:
        return self.model.predict(x)


class MyCustomModelComponent(ModelComponent):
    def __init__(self, model_path: str):
        self._model_path = model_path
        self.model = None

    def load(self):
        load_model = flama.load(self._model_path)
        self.model = MyCustomModel(load_model.model, load_model.meta, load_model.artifacts)

    def reset(self):
        self.model = None

    def resolve(self) -> MyCustomModel:
        if not self.model:
            self.load()

        assert self.model
        return self.model


component = MyCustomModelComponent("sklearn_model.flm")
# component = MyCustomModelComponent("pytorch_model.flm")
# component = MyCustomModelComponent("tensorflow_model.flm")


class MyCustomModelResource(BaseModelResource[MyCustomModelComponent]):
    name = "custom_model"
    verbose_name = "Lazy-loaded ScikitLearn Model"
    component = component

    info = {
        "model_version": "1.0.0",
        "library_version": "1.0.2",
    }

    def _get_metadata(self):
        return {
            "metadata": {
                "built-in": {
                    "verbose_name": self._meta.verbose_name,
                    "name": self._meta.name,
                },
                "custom": {
                    **self.info,
                    "loaded": self.component.model is not None,
                    "date": datetime.now().date(),
                    "time": datetime.now().time(),
                },
            }
        }

    @resource_method("/unload/", methods=["GET"], name="unload-method")
    def unload(self):
        """
        tags:
            - Lazy-loaded ScikitLearn Model
        summary:
            Unload the model.
        """
        self.component.reset()
        return self._get_metadata()

    @resource_method("/metadata/", methods=["GET"], name="metadata-method")
    def metadata(self):
        """
        tags:
            - Lazy-loaded ScikitLearn Model
        summary:
            Get metadata info.
        """
        return self._get_metadata()


app = Flama(
    openapi={
        "info": {
            "title": "Flama ML",
            "version": "0.1.0",
            "description": "Machine learning API using Flama 🔥",
        }
    },
    docs="/docs/",
    components=[component],
)

app.models.add_model_resource(path="/model", resource=MyCustomModelComponent)

if __name__ == "__main__":
    flama.run(flama_app="__main__:app", server_host="0.0.0.0", server_port=8080, server_reload=True)



import typing as t

from flama.ddd.workers.base import BaseWorker

__all__ = ["Worker"]


class Worker(BaseWorker):
    """Worker that does not apply any specific behavior.

    A basic implementation of the worker class that does not apply any specific behavior.
    """

    async def set_up(self) -> None:
        """First step in starting a unit of work."""
        ...

    async def tear_down(self, *, rollback: bool = False) -> None:
        """Last step in ending a unit of work.

        :param rollback: If the unit of work should be rolled back.
        """
        ...

    async def repository_params(self) -> tuple[list[t.Any], dict[str, t.Any]]:
        """Get the parameters for initialising the repositories.

        :return: Parameters for initialising the repositories.
        """
        return [], {}

    async def commit(self) -> None:
        """Commit the unit of work."""
        ...

    async def rollback(self) -> None:
        """Rollback the unit of work."""
        ...



import pytest

from flama import exceptions
from flama.applications import Flama
from flama.client import Client
from flama.resources.crud import CRUDResource
from flama.resources.resource import Resource
from flama.resources.routing import ResourceRoute, resource_method
from flama.routing import Mount, Route
from flama.sqlalchemy import SQLAlchemyModule


class TestCaseResourceRoute:
    @pytest.fixture(scope="function")
    def app(self):
        return Flama(schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")})

    @pytest.fixture(scope="function")
    def resource(self, puppy_model, puppy_schema):
        class PuppyResource(CRUDResource):
            name = "puppy"
            model = puppy_model
            schema = puppy_schema

        return PuppyResource()

    def test_init(self, resource):
        resource_route = ResourceRoute(
            "/puppy/",
            resource,
            tags={
                "create": {"tag": "create"},
                "retrieve": {"tag": "retrieve"},
                "update": {"tag": "update"},
                "partial_update": {"tag": "partial-update"},
                "delete": {"tag": "delete"},
                "list": {"tag": "list"},
                "replace": {"tag": "replace"},
                "partial_replace": {"tag": "partial-replace"},
                "drop": {"tag": "drop"},
            },
        )

        assert resource_route.path == "/puppy/"
        assert resource_route.resource == resource
        for route in resource_route.routes:
            assert isinstance(route, Route)
        assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
            ("/", {"POST"}, resource_route.resource.create, {"tag": "create"}),
            ("/{resource_id}/", {"GET", "HEAD"}, resource_route.resource.retrieve, {"tag": "retrieve"}),
            ("/{resource_id}/", {"PUT"}, resource_route.resource.update, {"tag": "update"}),
            ("/{resource_id}/", {"PATCH"}, resource_route.resource.partial_update, {"tag": "partial-update"}),
            ("/{resource_id}/", {"DELETE"}, resource_route.resource.delete, {"tag": "delete"}),
            ("/", {"GET", "HEAD"}, resource_route.resource.list, {"tag": "list"}),
            ("/", {"PUT"}, resource_route.resource.replace, {"tag": "replace"}),
            ("/", {"PATCH"}, resource_route.resource.partial_replace, {"tag": "partial-replace"}),
            ("/", {"DELETE"}, resource_route.resource.drop, {"tag": "drop"}),
        ]

    def test_init_wrong_tags(self, resource):
        with pytest.raises(exceptions.ApplicationError, match="Tags must be defined only for existing routes."):
            ResourceRoute(
                "/puppy/",
                resource,
                tags={
                    "create": {"tag": "create"},
                    "retrieve": {"tag": "retrieve"},
                    "update": {"tag": "update"},
                    "partial_update": {"tag": "partial-update"},
                    "delete": {"tag": "delete"},
                    "list": {"tag": "list"},
                    "replace": {"tag": "replace"},
                    "partial_replace": {"tag": "partial-replace"},
                    "drop": {"tag": "drop"},
                    "wrong": "wrong",
                },
            )

    def test_mount_resource_declarative(self, resource):
        routes = [Route("/", lambda: {"Hello": "world"}), ResourceRoute("/puppy/", resource)]

        app = Flama(routes=routes, schema=None, docs=None)

        assert len(app.router.routes) == 2

        assert len(app.routes) == 2
        resource_route = app.routes[1]
        assert isinstance(resource_route, ResourceRoute)
        assert [(route.path, route.methods, route.endpoint) for route in resource_route.routes] == [
            ("/", {"POST"}, resource_route.resource.create),
            ("/{resource_id}/", {"GET", "HEAD"}, resource_route.resource.retrieve),
            ("/{resource_id}/", {"PUT"}, resource_route.resource.update),
            ("/{resource_id}/", {"PATCH"}, resource_route.resource.partial_update),
            ("/{resource_id}/", {"DELETE"}, resource_route.resource.delete),
            ("/", {"GET", "HEAD"}, resource_route.resource.list),
            ("/", {"PUT"}, resource_route.resource.replace),
            ("/", {"PATCH"}, resource_route.resource.partial_replace),
            ("/", {"DELETE"}, resource_route.resource.drop),
        ]

    def test_nested_mount_resource(self, resource):
        app = Flama(schema=None, docs=None)
        app.add_route(route=Route("/", lambda: {"Hello": "world"}))

        sub_app = Flama(schema=None, docs=None)
        sub_app.resources.add_resource("/puppy/", resource)
        app.mount("/", sub_app)

        assert len(app.router.routes) == 2

        assert len(app.routes) == 2
        mount = app.routes[1]
        assert isinstance(mount, Mount)
        assert len(mount.routes) == 1
        resource_route = mount.routes[0]
        assert isinstance(resource_route, ResourceRoute)
        assert [(route.path, route.methods, route.endpoint) for route in resource_route.routes] == [
            ("/", {"POST"}, resource_route.resource.create),
            ("/{resource_id}/", {"GET", "HEAD"}, resource_route.resource.retrieve),
            ("/{resource_id}/", {"PUT"}, resource_route.resource.update),
            ("/{resource_id}/", {"PATCH"}, resource_route.resource.partial_update),
            ("/{resource_id}/", {"DELETE"}, resource_route.resource.delete),
            ("/", {"GET", "HEAD"}, resource_route.resource.list),
            ("/", {"PUT"}, resource_route.resource.replace),
            ("/", {"PATCH"}, resource_route.resource.partial_replace),
            ("/", {"DELETE"}, resource_route.resource.drop),
        ]

    async def test_request_nested_resource(self, app):
        class PuppyResource(Resource):
            name = "puppy"
            verbose_name = "Puppy"

            @resource_method("/", methods=["GET"], name="puppy-list", tags={"foo": "bar"})
            async def list(self):
                return {"name": "Canna"}

        sub_app = Flama(schema=None, docs=None)
        sub_app.resources.add_resource("/puppy/", PuppyResource)
        app.mount("/", sub_app)
        app.mark = 1
        sub_app.mark = 2

        async with Client(app) as client:
            response = await client.get("/puppy/")
            assert response.status_code == 200


class TestCaseResourceMethod:
    def test_resource_method(self):
        @resource_method(path="/", methods=["POST"], name="foo", tags={"additional": "bar"})
        def foo(x: int):
            return x

        assert hasattr(foo, "_meta")
        assert foo._meta.path == "/"
        assert foo._meta.methods == {"POST"}
        assert foo._meta.name == "foo"
        assert foo._meta.tags == {"additional": "bar"}
